from kqueen.config import current_config
from prometheus_client import multiprocess

import multiprocessing
import os

app_config = current_config()

bind = "{host}:{port}".format(
    host=app_config.get('KQUEEN_HOST'),
    port=app_config.get('KQUEEN_PORT'),
)
workers = multiprocessing.cpu_count() * 2 + 1

# check for prometheus settings
if 'prometheus_multiproc_dir' not in os.environ:
    raise Exception('Variable prometheus_multiproc_dir is required')


def child_exit(server, worker):
    multiprocess.mark_process_dead(worker.pid)



from flask import abort
from flask import current_app
from flask import jsonify
from flask import request
from flask.views import View
from flask_jwt import _jwt_required, current_identity, JWTError
from kqueen.auth import is_authorized
from .helpers import get_object


class GenericView(View):
    obj = None

    def get_class(self):
        if hasattr(self, 'object_class'):
            return self.object_class
        else:
            raise NotImplementedError('Missing object_class variable or get_class')

    def get_content(self, *args, **kwargs):
        raise NotImplementedError

    def check_authentication(self):
        _jwt_required(current_app.config['JWT_DEFAULT_REALM'])

    def check_authorization(self):
        # get view class
        try:
            _class = self.get_class()
        except NotImplementedError:
            return False

        # get user data
        if current_identity:
            user = current_identity.get_dict()
            organization = current_identity.organization
        else:
            return False

        # form policy key
        policy = '{}:{}'.format(_class.__name__.lower(), self.action)
        # get policies and update them with organization level overrides
        policies = current_app.config.get('DEFAULT_POLICIES', {})
        if hasattr(organization, 'policy') and organization.policy:
            policies.update(organization.policy)

        try:
            policy_value = policies[policy]
        except KeyError:
            current_app.logger.error('Unknown policy {}'.format(policy))
            return False

        # evaluate user permissions
        # if there are multiple objects, filter out those which current user
        # doesn't have access to
        if isinstance(self.obj, list):
            allowed = []
            for obj in self.obj:
                if is_authorized(user, policy_value, resource=obj):
                    allowed.append(obj)
            self.obj = allowed
        # if there is single object raise if user doesn't have access to it
        else:
            if not is_authorized(user, policy_value, resource=self.obj):
                raise JWTError('Insufficient permissions',
                               'Your user account is lacking the necessary '
                               'permissions to perform this operation')

    def set_object(self, *args, **kwargs):
        self.obj = get_object(self.get_class(), kwargs['pk'], current_identity)

        # check authorization for given object
        self.check_authorization()

    def dispatch_request(self, *args, **kwargs):
        self.check_authentication()
        self.set_object(*args, **kwargs)
        output = self.get_content(*args, **kwargs)

        return jsonify(output)


class GetView(GenericView):
    methods = ['GET']
    action = 'get'

    def get_content(self, *args, **kwargs):
        return self.obj


class DeleteView(GenericView):
    methods = ['DELETE']
    action = 'delete'

    def dispatch_request(self, *args, **kwargs):
        self.check_authentication()
        self.set_object(*args, **kwargs)

        try:
            self.obj.delete()
        except Exception:
            abort(500)

        return jsonify({'id': self.obj.id, 'state': 'deleted'})


class UpdateView(GenericView):
    methods = ['PATCH']
    action = 'update'

    def get_content(self, *args, **kwargs):
        return self.obj

    def dispatch_request(self, *args, **kwargs):
        self.check_authentication()

        if not request.json:
            abort(400, description='JSON data expected')

        data = request.json
        if not isinstance(data, dict):
            abort(400)

        self.set_object(*args, **kwargs)

        for key, value in data.items():
            setattr(self.obj, key, value)

        try:
            self.obj.save()
        except Exception:
            abort(500)

        output = self.get_content(*args, **kwargs)
        return jsonify(output)


class ListView(GenericView):
    methods = ['GET']
    action = 'list'

    def set_object(self, *args, **kwargs):
        try:
            namespace = current_identity.namespace
        except AttributeError:
            namespace = None

        self.obj = list(self.get_class().list(namespace, return_objects=True).values())
        self.check_authorization()

    def get_content(self, *args, **kwargs):
        return self.obj


class CreateView(GenericView):
    methods = ['POST']
    action = 'create'

    def save_object(self):
        self.obj.save()

    def after_save(self):
        pass

    def set_object(self, *args, **kwargs):
        cls = self.get_class()
        try:
            namespace = current_identity.namespace
        except AttributeError:
            namespace = None

        self.obj = cls.create(namespace, **request.json)
        self.check_authorization()

    def get_content(self, *args, **kwargs):
        return self.obj.get_dict(expand=True)

    def dispatch_request(self, *args, **kwargs):
        self.check_authentication()

        if not request.json:
            abort(400, description='JSON data expected')
        else:
            self.set_object(*args, **kwargs)
            try:
                self.save_object()
                self.after_save()
            except Exception as e:
                current_app.logger.error(e)
                abort(500, description='Creation failed with: {}'.format(e))

            output = self.get_content(*args, **kwargs)
            return jsonify(output)



"""Helpers are providing elementary functions and wrappers."""
import re

BINARY_PREFIX = {
    # decimal
    'm': 1000 ** -1,
    'k': 1000,
    'M': 1000 ** 2,
    'G': 1000 ** 3,
    'T': 1000 ** 4,
    'P': 1000 ** 5,
    # binary
    'mi': 1024 ** -1,
    'Ki': 1024,
    'Mi': 1024 ** 2,
    'Gi': 1024 ** 3,
    'Ti': 1024 ** 4,
    'Pi': 1024 ** 5,
}


def prefix_to_num(st):
    """Read string with prefix and return number.

    Args:
        st (string): String representation of value with prefix

    Returns:
        float: Calculated value without binary prefix

    Example:

        >>> prefix_to_num('1k')
        1000.0

    """
    num = ''
    prefix = ''

    # split number and prefix
    for i in range(len(st)):
        symbol = st[i]
        if symbol.isdigit() or symbol == '.':
            num += symbol
        else:
            prefix += symbol
    num = float(num)
    prefix = str(prefix).strip()

    # no prefix
    if prefix == '':
        return num

    # find multiplicator
    if prefix not in BINARY_PREFIX:
        raise ValueError('Prefix {} can not be parsed'.format(prefix))

    return num * BINARY_PREFIX[prefix]


def camel_split(st):
    """Split string by CamelCase.

    Args:
        st (string): Input string

    Returns:
        list: List of works splitted by camel case
    """

    return re.sub('(?!^)([A-Z][a-z]+)', r' \1', st).split()



from .test_crud import BaseTestCRUD
from flask import url_for
from kqueen.conftest import user
from kqueen.config import current_config

import bcrypt
import json
import pytest

config = current_config()


class TestUserCRUD(BaseTestCRUD):
    def get_object(self):
        return user()

    def get_edit_data(self):
        return {
            'username': 'patched user',
            'email': 'root@localhost',
        }

    def get_create_data(self):
        data = self.obj.get_dict()
        data['id'] = None
        data['organization'] = 'Organization:{}'.format(self.obj.organization.id)

        return data

    def get_auth_headers(self):
        """
        Get auth token for user in self.obj

        Returns:
            dict: Authorization header dict
        """
        data = {'username': self.obj.username, 'password': self.obj.username + 'password'}

        response = self.client.post(
            '/api/v1/auth',
            data=json.dumps(data),
            content_type='application/json')

        return {'Authorization': '{header_prefix} {token}'.format(
            header_prefix=config.get('JWT_AUTH_HEADER_PREFIX'),
            token=response.json['access_token'],
        )}

    def test_whoami(self):
        url = url_for('api.user_whoami')

        response = self.client.get(
            url,
            headers=self.get_auth_headers(),
            content_type='application/json',
        )

        assert response.json == self.obj.get_dict(expand=True)

    def test_namespace(self):
        user = self.get_object()

        assert user.namespace == user.organization.namespace

    def test_password_update(self):
        data = {'password': 'password123'}
        url = url_for('api.user_password_update', pk=self.obj.id)

        response = self.client.patch(
            url,
            data=json.dumps(data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 200
        patched = self.obj.__class__.load(
            self.namespace,
            self.obj.id,
        )
        user_password = patched.password.encode('utf-8')
        given_password = "password123".encode('utf-8')
        assert bcrypt.checkpw(given_password, user_password)

    def test_crud_update(self):
        data = self.get_edit_data()

        response = self.client.patch(
            self.urls['update'],
            data=json.dumps(data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 200
        patched = self.obj.__class__.load(
            self.namespace,
            self.obj.id,
        )
        for key, value in data.items():
            assert getattr(patched, key) == value

    @pytest.mark.last
    def test_crud_delete(self):
        super(TestUserCRUD, self).test_crud_delete()



from .base import BaseConfig


class Config(BaseConfig):
    DEBUG = True
    LOG_LEVEL = 'DEBUG'

    # App secret
    SECRET_KEY = 'secret'

    # Jenkins engine settings
    JENKINS_API_URL = 'https://ci.mcp.mirantis.net'
    JENKINS_PROVISION_JOB_NAME = 'deploy-aws-k8s_ha_calico_sm'
    JENKINS_PROVISION_JOB_CTX = {}
    JENKINS_ANCHOR_PARAMETER = 'STACK_NAME'
    JENKINS_USERNAME = None
    JENKINS_PASSWORD = None



from flask import current_app
from flask import url_for

import pytest


@pytest.mark.usefixtures('client_class')
class TestMetrics:
    def setup(self):
        self.url = url_for('metrics.root')
        self.response = self.client.get(self.url)
        self.content = self.response.data.decode(self.response.charset)

    def test_assert_status(self):
        assert self.response.status_code == 200


class TestAccess:
    def setup(self):
        self.url = url_for('metrics.root')

    @pytest.mark.parametrize('whitelist,code', [
        ('127.0.0.0/8', 200),
        ('8.8.8.8/32', 401),
        ('2001:db8:a0b:12f0::/64', 401),
    ])
    def test_metrics_acccess(self, client, monkeypatch, whitelist, code):
        monkeypatch.setitem(current_app.config, 'PROMETHEUS_WHITELIST', whitelist)

        response = client.get(self.url)
        assert response.status_code == code



from flask import request
from prometheus_client import Counter
from prometheus_client import Histogram

import os
import time

# Prometheus metrics
REQUEST_COUNT = Counter(
    'request_count',
    'HTPP Request Count',
    ['method', 'endpoint', 'http_status']
)
REQUEST_LATENCY = Histogram(
    'request_latency',
    'HTTP Request latency',
    ['method', 'endpoint']
)


def start_timer():
    request.start_time = time.time()


def record_request_data(response):
    resp_time = time.time() - request.start_time

    REQUEST_LATENCY.labels(request.method, request.path).observe(resp_time)
    REQUEST_COUNT.labels(request.method, request.path, response.status_code).inc()

    return response


def check_prometheus():
    is_gunicorn = "gunicorn" in os.environ.get("SERVER_SOFTWARE", "")

    if 'prometheus_multiproc_dir' in os.environ:
        os.makedirs(os.environ['prometheus_multiproc_dir'], exist_ok=True)

    elif is_gunicorn:
        raise Exception('Please set prometheus_multiproc_dir variable using `export prometheus_multiproc_dir=$(mktemp -d)`')


def setup_metrics(app):

    check_prometheus()

    app.before_request(start_timer)
    app.after_request(record_request_data)



from .test_crud import BaseTestCRUD
from flask import url_for
from kqueen.config import current_config
from kqueen.conftest import organization

config = current_config()


class TestOrganizationCRUD(BaseTestCRUD):
    def get_object(self):
        return organization()

    def get_edit_data(self):
        return {
            'name': 'patched organization',
            'namespace': 'namespace123',
        }

    def test_policy(self):
        url = url_for('api.organization_policy', pk=self.obj.id)

        policies = config.get('DEFAULT_POLICIES', {})
        if hasattr(self.obj, 'policy') and self.obj.policy:
            policies.update(self.obj.policy)

        response = self.client.get(
            url,
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 200
        assert response.json == policies



from flask import url_for
from kqueen.conftest import app
from uuid import uuid4

import pytest

methods = ['GET', 'POST', 'PATCH', 'DELETE']
url_skip = ['/api/v1/health']


def generate_arg(name):
    """Return sample argument value for argument name."""

    if name == 'pk':
        return uuid4()

    elif name == 'filename':
        return 'test.json'
    else:
        raise KeyError('Argument name {} not known'.format(name))


def skip_rule(rule):
    return not rule.endpoint.startswith('api.')


def get_urls():
    """List all used urls, including pks."""
    app_instance = app()
    app_instance.config['SERVER_NAME'] = 'server'
    urls = []

    for rule in app_instance.url_map.iter_rules():
        if skip_rule(rule):
            continue

        options = {'_external': False}
        for arg_name in rule.arguments:
            options[arg_name] = generate_arg(arg_name)

        with app_instance.app_context():
            url = url_for(rule.endpoint, **options)
            if url not in url_skip:
                urls.append(url)

    return urls


def test_root(client, auth_header):
    response = client.get(url_for('api.index'), headers=auth_header)

    assert response.json == {'response': 'Gutten tag!'}


# TODO: fix bad request code
@pytest.mark.skip('not working yet')
@pytest.mark.parametrize('method_name', methods)
@pytest.mark.parametrize('url', get_urls())
def test_all_views_require_token(client, method_name, url):
    method = getattr(client, method_name.lower())
    response = method(
        url,
        data='{}',
        content_type='application/json',
    )
    assert response.status_code in {401, 405}



from flask import abort
from uuid import UUID


def get_object(object_class, pk, user=None):
    # read uuid
    if isinstance(pk, UUID):
        object_id = pk
    else:
        try:
            object_id = UUID(pk, version=4)
        except ValueError:
            abort(400)

    # read namespace for user
    try:
        namespace = user.namespace
    except AttributeError:
        namespace = None

    # load object
    try:
        obj = object_class.load(namespace, object_id)
    except NameError:
        abort(404)

    return obj



from kqueen.config import current_config
from kqueen.engines.base import BaseEngine

from azure.common.credentials import ServicePrincipalCredentials
from azure.mgmt.containerservice import ContainerServiceClient

import logging
import base64
import yaml

logger = logging.getLogger(__name__)
config = current_config()

STATE_MAP = {
    'Creating': config.get('CLUSTER_PROVISIONING_STATE'),
    'Succeeded': config.get('CLUSTER_OK_STATE'),
    'Deleting': config.get('CLUSTER_DEPROVISIONING_STATE'),
    'Failed': config.get('CLUSTER_ERROR_STATE')
}


class AksEngine(BaseEngine):
    """
    Azure Container Service
    """
    name = 'aks'
    verbose_name = 'Azure Kubernetes Managed Service'
    resource_group_name = 'test-cluster'
    parameter_schema = {
        'provisioner': {
            'client_id': {
                'type': 'text',
                'label': 'Client ID',
                'validators': {
                    'required': True
                }
            },
            'secret': {
                'type': 'password',
                'label': 'Secret',
                'validators': {
                    'required': True
                }
            },
            'tenant': {
                'type': 'text',
                'label': 'Tenant',
                'validators': {
                    'required': True
                }
            },
            'subscription_id': {
                'type': 'text',
                'label': 'Subscription ID',
                'validators': {
                    'required': True
                }
            },
        },
        'cluster': {
            'location': {
                'type': 'select',
                'label': 'Location',
                'choices': [
                    ('eastus', 'East US'),
                    ('centralus', 'Central US'),
                    ('westeurope', 'West Europe')
                ],
                'validators': {
                    'required': True
                }
            },
            'ssh_key': {
                'type': 'text_area',
                'label': 'SSH Key (public)',
                'validators': {}
            },
            'node_count': {
                'type': 'integer',
                'label': 'Node Count',
                'default': 1,
                'validators': {
                    'required': True,
                    'min': 1,
                    'number': True
                }
            }
        }
    }

    def __init__(self, cluster, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.__init__`
        """
        # Call parent init to save cluster on self
        super(AksEngine, self).__init__(cluster, **kwargs)

        # Client initialization
        self.client_id = kwargs.get('client_id', '')
        self.secret = kwargs.get('secret', '')
        self.tenant = kwargs.get('tenant', '')
        self.subscription_id = kwargs.get('subscription_id', '')
        self.resource_group_name = kwargs.get('resource_group_name', self.resource_group_name)
        self.location = kwargs.get('location', '')
        self.ssh_key = kwargs.get('ssh_key', '')
        self.node_count = kwargs.get('node_count', 1)
        self.client = self._get_client()

        # Cache settings
        self.cache_timeout = 5 * 60

    def _get_client(self):
        """
        Initialize Azure client
        Construct service account credentials using the service account key file

        """
        credentials = ServicePrincipalCredentials(client_id=self.client_id, secret=self.secret, tenant=self.tenant)
        subscription_id = self.subscription_id
        client = ContainerServiceClient(credentials, subscription_id)

        return client

    def provision(self, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.provision`
        """
        cluster = {
            'location': self.location,
            'type': 'Microsoft.ContainerService/ManagedClusters',
            'name': self.cluster.id,
            'properties': {
                # TODO: fix hardcoded params
                'kubernetes_version': '1.7.7',
                'dns_prefix': 'test-cluster',
                'agent_pool_profiles': [
                    {
                        'fqdn': None,
                        'vnet_subnet_id': None,
                        'storage_profile': 'ManagedDisks',
                        'name': 'agentpool',
                        'count': self.node_count,
                        'dns_prefix': None,
                        'ports': None,
                        'vm_size': 'Standard_D2_v2',
                        'os_type': 'Linux',
                        'os_disk_size_gb': None
                    }
                ],
                'service_principal_profile': {
                    'client_id': self.client_id,
                    'secret': self.secret
                },
                'linux_profile': {
                    'admin_username': 'azureuser',
                    'ssh': {
                        'public_keys': [
                            {
                                'key_data': self.ssh_key
                            }
                        ]
                    }
                }
            }
        }

        try:
            self.client.managed_clusters.create_or_update(self.resource_group_name, self.cluster.id, cluster)
            # TODO: check if provisioning response is healthy
        except Exception as e:
            msg = 'Creating cluster {} failed with following reason: {}'.format(self.cluster.id, repr(e))
            logger.error(msg)
            return False, msg

        return True, None

    def deprovision(self, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.deprovision`
        """
        try:
            self.client.managed_clusters.delete(self.resource_group_name, self.cluster.id)
            # TODO: check if deprovisioning response is healthy
        except Exception as e:
            msg = 'Deleting cluster {} failed with following reason: {}'.format(self.cluster.id, repr(e))
            logger.error(msg)
            return False, msg

        return True, None

    def get_kubeconfig(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.get_kubeconfig`
        """
        if not self.cluster.kubeconfig:
            cluster = self.client.managed_clusters.get(self.resource_group_name, self.cluster.id)

            kubeconfig = {}

            if cluster.properties.provisioning_state != "Succeeded":
                return self.cluster.kubeconfig

            access_profiles = cluster.properties.access_profiles.as_dict()
            access_profile = access_profiles.get('cluster_admin')
            encoded_kubeconfig = access_profile.get("kube_config")
            kubeconfig = base64.b64decode(encoded_kubeconfig).decode(encoding='UTF-8')

            self.cluster.kubeconfig = yaml.load(kubeconfig)
            self.cluster.save()

        return self.cluster.kubeconfig

    def cluster_get(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.cluster_get`

        First we try to get cluster by external_id, because its much more efficient in this
        implementation. If its not possible yet, we return from the slower method
        """
        try:
            response = self.client.managed_clusters.get(self.resource_group_name, self.cluster.id)
        except Exception as e:
            msg = 'Fetching data from backend for cluster {} failed with following reason: {}'.format(self.cluster.id, repr(e))
            logger.error(msg)
            return {}
        properties = response.properties.as_dict()
        state = STATE_MAP.get(properties.get('provisioning_state'), config.get('CLUSTER_UNKNOWN_STATE'))

        key = 'cluster-{}-{}'.format(self.name, self.cluster.id)
        cluster = {
            'key': key,
            'name': self.cluster.id,
            'id': self.cluster.id,
            'state': state,
            'metadata': {}
        }
        return cluster

    def cluster_list(self):
        """AKS engine don't support list of clusters"""
        # TODO: it does, add list of clusters

        return []



import importlib
import logging
import os

CONFIG_FILE_DEFAULT = 'config/dev.py'
logger = logging.getLogger(__name__)


def select_file(config_file=None):
    """
    Select file to be used as a configuration.

    Attributes:
        config_file (str): Filename to be used as a configuration file

    Returns:
        str: filename to be used as a configuration file
    """

    if not config_file or config_file == 'None':
        config_file = os.environ.get('KQUEEN_CONFIG_FILE')
        logger.debug('Config file from env variable: {}'.format(config_file))

    if not config_file or config_file == 'None':
        config_file = CONFIG_FILE_DEFAULT
        logger.debug('Config file, using default: {}'.format(config_file))

    return config_file


def apply_env_changes(config, prefix='KQUEEN_'):
    """
    Read env variables starting with prefix and apply
    them to existing configuration

    Attributes:
        config (obj): Connfiguration object. This configuration will updated.
        prefix (str): Prefix for environment variables. Defaults to `KQUEEN_`.

    """

    for name, value in os.environ.items():
        if name.startswith(prefix):
            config_key_name = name[len(prefix):]
            setattr(config, config_key_name, value)


def current_config(config_file=None):
    read_file = select_file(config_file)
    logger.debug('Loading config from {}'.format(read_file))

    module_name = read_file.replace('/', '.').replace('.py', '')

    module = importlib.import_module('kqueen.{}'.format(module_name))
    config = getattr(module, 'Config')
    apply_env_changes(config)

    setattr(config, 'source_file', read_file)

    return config



from flask import url_for
from kqueen.conftest import auth_header, user_with_namespace, get_auth_token
from kqueen.config import current_config

import json
import pytest

config = current_config()


@pytest.mark.usefixtures('client_class')
class BaseTestCRUD:
    def get_object(self):
        raise NotImplementedError

    def get_edit_data(self):
        raise NotImplementedError

    def get_resource_type(self):
        return self.obj.__class__.__name__.lower()

    def get_create_data(self):
        data = self.obj.get_dict(expand=True)
        data['id'] = None

        return data

    def get_urls(self, pk=None):

        if not pk:
            pk = self.obj.id

        return {
            'list': url_for(
                'api.{}_list'.format(self.get_resource_type())
            ),
            'create': url_for(
                'api.{}_create'.format(self.get_resource_type())
            ),
            'get': url_for(
                'api.{}_get'.format(self.get_resource_type()),
                pk=pk
            ),
            'update': url_for(
                'api.{}_update'.format(self.get_resource_type()),
                pk=pk
            ),
            'delete': url_for(
                'api.{}_delete'.format(self.get_resource_type()),
                pk=pk
            ),
        }

    def setup(self):
        self.obj = self.get_object()
        self.obj.save()

        self.auth_header = auth_header(self.client)
        self.namespace = self.auth_header['X-Test-Namespace']
        self.urls = self.get_urls()

    def test_crud_create(self):
        data = self.get_create_data()

        response = self.client.post(
            self.urls['create'],
            data=json.dumps(data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 200
        assert response.json['id'] != self.obj.id

    def test_crud_create_not_json(self):
        response = self.client.post(
            self.urls['create'],
            headers=self.auth_header,
            content_type='text/plain',
        )

        assert response.status_code == 400

    def test_crud_create_failed_save(self, monkeypatch):
        def fake_save(self, *args, **kwargs):
            raise Exception('Test')

        monkeypatch.setattr(self.obj.__class__, 'save', fake_save)
        data = self.get_create_data()

        response = self.client.post(
            self.urls['create'],
            data=json.dumps(data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 500

    def test_crud_get(self):
        self.obj.save()

        response = self.client.get(
            self.urls['get'],
            headers=self.auth_header,
        )

        assert response.status_code == 200
        assert response.json == self.obj.get_dict(expand=True)

    def test_crud_list(self):
        response = self.client.get(
            self.urls['list'],
            headers=self.auth_header,
        )

        data = response.json

        assert isinstance(data, list)
        assert len(data) == len(self.obj.__class__.list(
            self.namespace,
            return_objects=False)
        )
        assert self.obj.get_dict(expand=True) in data

    def test_crud_update(self):
        data = self.get_edit_data()

        response = self.client.patch(
            self.urls['update'],
            data=json.dumps(data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 200
        patched = self.obj.__class__.load(
            self.namespace,
            self.obj.id,
        )
        for key, value in data.items():
            assert getattr(patched, key) == value

    @pytest.mark.parametrize('data,content_type,code', [
        ('hello', 'text/plain', 400),
        ('[1, 2, 3]', 'application/json', 400),
    ])
    def test_crud_update_return_error(self, data, content_type, code):
        response = self.client.patch(
            self.urls['update'],
            data=data,
            headers=self.auth_header,
            content_type=content_type,
        )

        assert response.status_code == code

    def test_crud_update_failed_save(self, monkeypatch):
        def fake_save(self, *args, **kwargs):
            raise Exception('Testing')

        monkeypatch.setattr(self.obj.__class__, 'save', fake_save)

        data = self.get_edit_data()
        response = self.client.patch(
            self.urls['update'],
            data=json.dumps(data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 500

    def test_crud_delete(self):
        response = self.client.delete(
            self.urls['delete'],
            headers=self.auth_header,
        )

        assert response.status_code == 200
        with pytest.raises(NameError, message='Object not found'):
            self.obj.__class__.load(
                self.namespace,
                self.obj.id,
            )

    def test_crud_delete_failed(self, monkeypatch):
        def fake_delete(self, *args, **kwargs):
            raise Exception('Testing')

        monkeypatch.setattr(self.obj.__class__, 'delete', fake_delete)

        response = self.client.delete(
            self.urls['delete'],
            headers=self.auth_header,
        )

        assert response.status_code == 500

    #
    # namespacing tests
    #
    @pytest.fixture
    def setup_namespace(self):
        self.user1 = user_with_namespace()
        self.user2 = user_with_namespace()

    @pytest.mark.usefixtures('setup_namespace')
    def test_namespacing(self, client):
        obj = self.get_object()

        # skip if object class isn't namespaced
        if not obj.__class__.is_namespaced():
            pytest.skip('Class {} isn\'t namespaced'.format(obj.__class__.__name__))

        objs = {}

        # create objects for both users
        for u in [self.user1, self.user2]:
            data = self.get_create_data()

            auth_header = get_auth_token(self.client, u)
            headers = {
                'Authorization': '{} {}'.format(
                    config.get('JWT_AUTH_HEADER_PREFIX'),
                    auth_header
                )
            }

            # TODO: fix this
            # Dirty hack to make testing data namespaced.
            organization_data = {
                'name': 'Test organization',
                'namespace': 'testorg',
            }
            response = self.client.post(
                url_for('api.organization_create'),
                data=json.dumps(organization_data),
                headers=headers,
                content_type='application/json',
            )
            organization_ref = 'Organization:{}'.format(response.json['id'])
            owner_data = {
                'username': 'Test owner',
                'email': 'owner@pytest.org',
                'password': 'pytest',
                'organization': organization_ref,
                'role': 'admin',
                'active': True
            }
            response = self.client.post(
                url_for('api.user_create'),
                data=json.dumps(owner_data),
                headers=headers,
                content_type='application/json',
            )
            if 'owner' in data:
                data['owner'] = 'User:{}'.format(response.json['id'])
            if 'provisioner' in data:
                provisioner_data = {
                    'name': 'Test provisioner',
                    'engine': 'kqueen.engines.ManualEngine',
                    'owner': 'User:{}'.format(response.json['id'])
                }
                response = self.client.post(
                    url_for('api.provisioner_create'),
                    data=json.dumps(provisioner_data),
                    headers=headers,
                    content_type='application/json',
                )
                data['provisioner'] = 'Provisioner:{}'.format(response.json['id'])

            response = self.client.post(
                self.urls['create'],
                data=json.dumps(data),
                headers=headers,
                content_type='application/json',
            )

            print(response.data.decode(response.charset))
            objs[u.namespace] = response.json['id']

            print(response.json)

        # test use can't read other's object
        for u in [self.user1, self.user2]:
            auth_header = get_auth_token(self.client, u)
            headers = {
                'Authorization': '{} {}'.format(
                    config.get('JWT_AUTH_HEADER_PREFIX'),
                    auth_header
                )
            }

            for ns, pk in objs.items():
                url = self.get_urls(pk)['get']

                if ns == u.namespace:
                    req_code = 200
                else:
                    req_code = 404

                response = self.client.get(
                    url,
                    headers=headers,
                    content_type='application/json',
                )

                print(response.data.decode(response.charset))
                assert response.status_code == req_code



from .jenkins import JenkinsEngine


class TestJenkinsInit:
    def setup(self):
        pass

    def test_init_jenkins(self, cluster):
        JenkinsEngine(cluster)

    def test_init_jenkins_with_kwargs(self, cluster):
        engine = JenkinsEngine(cluster, username='foo', password='bar')
        assert engine.username == 'foo'
        assert engine.password == 'bar'



from .auth import authenticate
from .auth import identity
from .blueprints.api.views import api
from .blueprints.metrics.views import metrics
from .config import current_config
from .middleware import setup_metrics
from .serializers import KqueenJSONEncoder
from .storages.etcd import EtcdBackend
from flask import Flask
from flask_jwt import JWT
from flask_swagger_ui import get_swaggerui_blueprint
from werkzeug.contrib.cache import SimpleCache

import logging

logger = logging.getLogger(__name__)
cache = SimpleCache()
swagger_url = '/api/docs'
api_url = '/api/v1/swagger'

swaggerui_blueprint = get_swaggerui_blueprint(
    swagger_url,
    api_url,
    config={
        'docExpansion': 'list'
    }
)


def create_app(config_file=None):
    app = Flask(__name__, static_folder='./asset/static')
    app.json_encoder = KqueenJSONEncoder

    app.register_blueprint(api, url_prefix='/api/v1')
    app.register_blueprint(metrics, url_prefix='/metrics')
    app.register_blueprint(swaggerui_blueprint, url_prefix=swagger_url)

    # load configuration
    config = current_config(config_file)
    config.setup_policies()
    app.config.from_mapping(config.to_dict())
    app.logger.setLevel(getattr(logging, app.config.get('LOG_LEVEL')))
    app.logger.info('Loading configuration from {}'.format(config.source_file))

    # setup database
    app.db = EtcdBackend()

    # setup JWT
    JWT(app, authenticate, identity)

    # setup metrics
    setup_metrics(app)

    return app


app = create_app()


def run():
    logger.debug('kqueen starting')
    app.run(
        host=app.config.get('KQUEEN_HOST'),
        port=int(app.config.get('KQUEEN_PORT'))
    )



from .utils import current_config
from .utils import select_file
from kqueen.server import create_app
from .base import BaseConfig

import pytest

config_envvar = 'KQUEEN_CONFIG_FILE'
config_file_default = 'config/test.py'

file_parametrization = [
    ('/tmp/test.py', '/tmp/test.py'),
    (None, config_file_default),
    ('', config_file_default),
]


class TestSelectConfig:
    @pytest.mark.parametrize('config_file,req', [
        ('/tmp/test.py', '/tmp/test.py'),
        (None, 'config/test.py'),
        ('', 'config/test.py'),
    ])
    def test_select_config(self, config_file, req):
        assert select_file(config_file) == req

    @pytest.mark.parametrize('config_file,req', [
        ('/tmp/test.py', '/tmp/test.py'),
        (None, 'config/dev.py'),
        ('', 'config/dev.py'),
    ])
    def test_select_config_from_envvar(self, monkeypatch, config_file, req):
        monkeypatch.setenv(config_envvar, config_file)
        assert select_file() == req


class TestCurrentConfig:
    @pytest.mark.parametrize('config_file', [
        'config/dev.py',
        'config/test.py',
    ])
    def test_current_config(self, config_file):
        config = current_config(config_file)

        assert config.source_file == select_file(config_file)


class TestConfigFromEnv:
    @pytest.mark.parametrize('name,value', [
        ('KQUEEN_DUMMY', '123'),
        ('KQUEEN_ETCD_HOST', '4001'),
    ])
    def test_env_var(self, monkeypatch, name, value):
        monkeypatch.setenv(name, value)
        config = current_config()

        config_key_name = name[7:]

        assert config.get(config_key_name) == value

    def test_env_var_in_app(self, monkeypatch):
        monkeypatch.setenv('KQUEEN_DUMMY', '123')

        app = create_app()

        assert app.config.get('DUMMY') == '123'


class TestBaseFunction:
    def setup(self):
        self.cl = BaseConfig

    def test_to_dict(self):
        dicted = self.cl.to_dict()

        assert isinstance(dicted, dict)
        assert dicted['DEBUG'] is False

    def test_get_regular(self):
        assert self.cl.get('DEBUG') is False

    def test_get_default(self):
        assert self.cl.get('NONEXISTING', 123) == 123



from datetime import timedelta

import json
import os


class BaseConfig:
    DEBUG = False
    LOG_LEVEL = 'WARNING'

    KQUEEN_HOST = '127.0.0.1'
    KQUEEN_PORT = 5000

    # etcd settings
    # ETCD_HOST
    # ETCD_PORT

    # JWT auth options
    JWT_DEFAULT_REALM = 'Login Required'
    JWT_AUTH_URL_RULE = '/api/v1/auth'
    JWT_EXPIRATION_DELTA = timedelta(hours=1)
    JWT_AUTH_HEADER_PREFIX = 'Bearer'

    BCRYPT_ROUNDS = 12

    # Cluster statuses
    CLUSTER_ERROR_STATE = 'Error'
    CLUSTER_OK_STATE = 'OK'
    CLUSTER_PROVISIONING_STATE = 'Deploying'
    CLUSTER_DEPROVISIONING_STATE = 'Destroying'
    CLUSTER_UNKNOWN_STATE = 'Unknown'

    CLUSTER_STATE_ON_LIST = True

    # Provisioner statuses
    PROVISIONER_ERROR_STATE = 'Error'
    PROVISIONER_OK_STATE = 'OK'
    PROVISIONER_UNKNOWN_STATE = 'Not Reachable'

    PROVISIONER_ENGINE_WHITELIST = None

    PROMETHEUS_WHITELIST = '127.0.0.0/8'

    @classmethod
    def get(cls, name, default=None):
        """Emulate get method from dict"""

        if hasattr(cls, name):
            return getattr(cls, name)
        else:
            return default

    @classmethod
    def to_dict(cls):
        """Return dict of all uppercase attributes"""

        out = {}

        for att_name in dir(cls):
            if att_name.isupper():
                out[att_name] = getattr(cls, att_name)

        return out

    @classmethod
    def setup_policies(cls):
        """Read default policy file"""

        base_path = os.path.dirname(__file__)
        full_path = os.path.join(base_path, 'default_policy.json')
        fh = open(full_path, 'r')
        try:
            policy_string = fh.read()
            cls.DEFAULT_POLICIES = json.loads(policy_string)
        except Exception:
            cls.DEFAULT_POLICIES = {}
        finally:
            fh.close()



from google.oauth2 import service_account
from kqueen.config import current_config
from kqueen.engines.base import BaseEngine

import googleapiclient.discovery
import logging
import requests

logger = logging.getLogger(__name__)
config = current_config()


STATE_MAP = {
    'PROVISIONING': config.get('CLUSTER_PROVISIONING_STATE'),
    'RUNNING': config.get('CLUSTER_OK_STATE'),
    'STOPPING': config.get('CLUSTER_DEPROVISIONING_STATE')
}


class GceEngine(BaseEngine):
    """
    Google Container Engine
    """
    name = 'gce'
    verbose_name = 'Google Container engine'
    # TODO: only subset of possible choices for zone is listed in parameter_schema,
    # we could add more later, here is the list of possible choices:
    # https://cloud.google.com/compute/docs/regions-zones/
    parameter_schema = {
        'provisioner': {
            'service_account_info': {
                'type': 'json_file',
                'label': 'Service Account File (JSON)',
                'validators': {
                    'required': True,
                    'jsonfile': [
                        'private_key_id',
                        'private_key',
                        'client_email',
                        'client_id',
                        'auth_uri',
                        'token_uri'
                    ]
                }
            },
            'project': {
                'type': 'text',
                'label': 'Project',
                'validators': {
                    'required': True,
                }
            }
        },
        'cluster': {
            'node_count': {
                'type': 'integer',
                'label': 'Node Count',
                'default': 1,
                'validators': {
                    'required': True,
                    'min': 1,
                    'number': True
                }
            },
            'zone': {
                'type': 'select',
                'label': 'Zone',
                'choices': [
                    ('us-central1-a', 'US - Central 1 - A'),
                    ('us-west1-a', 'US - West 1 - A'),
                    ('us-east1-a', 'US - East 1 - A'),
                    ('southamerica-east1-a', 'South America - East 1 - A'),
                    ('europe-west1-b', 'Europe - West 1 - B'),
                    ('asia-southeast1-a', 'Asia - Southeast 1 - A')
                ],
                'validators': {
                    'required': True
                }
            }
        }
    }

    def __init__(self, cluster, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.__init__`
        """
        # Call parent init to save cluster on self
        super(GceEngine, self).__init__(cluster, **kwargs)
        # Client initialization
        self.service_account_info = kwargs.get('service_account_info', {})
        self.project = kwargs.get('project', '')
        self.zone = kwargs.get('zone', '')
        self.cluster_id = 'a' + self.cluster.id.replace('-', '')
        self.cluster_config = {
            'cluster': {
                'name': self.cluster_id,
                'initialNodeCount': kwargs.get('node_count', 1)
            }
        }
        self.client = self._get_client()
        # Cache settings
        self.cache_timeout = 5 * 60

    def _get_client(self):
        """
        Initialize Google client
        Construct service account credentials using the service account key file

        """
        credentials = service_account.Credentials.from_service_account_info(self.service_account_info)
        client = googleapiclient.discovery.build('container', 'v1', credentials=credentials)

        return client

    def provision(self, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.provision`
        """
        try:
            self.client.projects().zones().clusters().create(projectId=self.project, zone=self.zone, body=self.cluster_config).execute()
            # TODO: check if provisioning response is healthy
        except Exception as e:
            msg = 'Creating cluster {} failed with following reason: {}'.format(self.cluster_id, repr(e))
            logger.error(msg)
            return False, msg

        return True, None

    def deprovision(self, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.deprovision`
        """
        try:
            self.client.projects().zones().clusters().delete(projectId=self.project, zone=self.zone, clusterId=self.cluster_id).execute()
            # TODO: check if provisioning response is healthy
        except Exception as e:
            msg = 'Deleting cluster {} failed with following reason: {}'.format(self.cluster_id, repr(e))
            logger.error(msg)
            return False, msg

        return True, None

    def get_kubeconfig(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.get_kubeconfig`
        """
        if not self.cluster.kubeconfig:
            cluster = self.client.projects().zones().clusters().get(projectId=self.project, zone=self.zone, clusterId=self.cluster_id).execute()

            kubeconfig = {}

            if cluster["status"] != "RUNNING":
                return self.cluster.kubeconfig

            # Get cluster section in KubeConfig
            kubeconfig_cluster_detail = {
                'server': "https://" + cluster["endpoint"],
                'certificate-authority-data': cluster["masterAuth"]["clusterCaCertificate"]
            }

            kubeconfig_cluster = {
                'name': cluster["name"],
                'cluster': kubeconfig_cluster_detail
            }

            user = {}
            user['username'] = cluster["masterAuth"]["username"]
            user['password'] = cluster["masterAuth"]["password"]

            kubeconfig_user = {
                'name': 'admin',
                'user': user
            }

            kubeconfig_context = {
                'name': cluster["name"],
                'context': {
                    'cluster': cluster["name"],
                    'user': kubeconfig_user["name"],
                },
            }

            kubeconfig = {
                'apiVersion': 'v1',
                'contexts': [kubeconfig_context],
                'clusters': [kubeconfig_cluster],
                'current-context': cluster["name"],
                'kind': 'Config',
                'preferences': {},
                'users': [kubeconfig_user],
            }

            self.cluster.kubeconfig = kubeconfig
            self.cluster.save()

        return self.cluster.kubeconfig

    def cluster_get(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.cluster_get`

        First we try to get cluster by external_id, because its much more efficient in this
        implementation. If its not possible yet, we return from the slower method
        """
        request = self.client.projects().zones().clusters().get(projectId=self.project, zone=self.zone, clusterId=self.cluster_id)
        try:
            response = request.execute()
        except Exception as e:
            msg = 'Fetching data from backend for cluster {} failed with following reason: {}'.format(self.cluster_id, repr(e))
            logger.error(msg)
            return {}

        state = STATE_MAP.get(response['status'], config.get('CLUSTER_UNKNOWN_STATE'))

        key = 'cluster-{}-{}'.format(self.name, self.cluster_id)
        cluster = {
            'key': key,
            'name': self.cluster_id,
            'id': self.cluster.id,
            'state': state,
            'metadata': {}
        }
        return cluster

    def cluster_list(self):
        """GCE engine don't support list of clusters"""

        return []

    @classmethod
    def engine_status(cls):
        test_url = 'https://container.googleapis.com/v1/projects/project/zones/zone/clusters?alt=json'
        headers = {'Accept': 'application/json'}
        response = requests.get(test_url, headers=headers)
        if response.status_code == 401:
            return config.get('PROVISIONER_OK_STATE')
        return config.get('PROVISIONER_ERROR_STATE')



from .base import BaseConfig


class Config(BaseConfig):
    DEBUG = True
    LOG_LEVEL = 'DEBUG'

    # App secret
    SECRET_KEY = 'secret'

    # Etcd settings
    ETCD_PREFIX = '/kqueen_test'

    BCRYPT_ROUNDS = 4

    # Jenkins engine settings
    JENKINS_API_URL = 'http://localhost'
    JENKINS_PROVISION_JOB_NAME = 'job_name'
    JENKINS_PROVISION_JOB_CTX = {}
    JENKINS_ANCHOR_PARAMETER = 'STACK_NAME'
    JENKINS_USERNAME = None
    JENKINS_PASSWORD = None

    # Azure Container Service
    AKS_CLIENT_ID = '9ce5569a-8207-4bwe-b2a7-fba6da19a162'
    AKS_SECRET = '24dee47b-451f-b8c2-aae2-8c22b5ce353a'
    AKS_TENANT = '9be2469a-4466-451f-b8c2-3c1156f933a5'
    AKS_SUBSCRIPTION_ID = 'b0f2aac6-a64a-45df-89ec-3c1156f933a5'

    # SSH public key
    SSH_KEY = 'ssh-rsa AAAAB3NzadfadfafQEAylDZDzgMuEsJQpwFHDW+QivCVhryxXd1/HWqq1TVhJmT9oNAYdhUBnf/9kVtgmP0EWpDJtGSEaSugCmx8KE76I64RhpOTlm7wO0FFUVnzhFtTPx38WHfMjMdk1HF8twZU4svi72Xbg1KyBimwvaxTTd4zxq8Mskp3uwtkqPcQJDSQaZYv+wtuB6m6vHBCOTZwAognDGEvvCg0dgTU4hch1zoHSaxedS1UFHjUAM598iuI3+hMos/5hjG/vuay4cPLBJX5x1YF6blbFALwrQw8ZmTPaimqDUA9WD6KSmS1qg4rOkk4cszIfJ5vyymMrG+G3qk5LeT4VrgIgWQTAHyXw=='



from .__init__ import __all__ as all_engines
from .base import BaseEngine
from .jenkins import JenkinsEngine

import pytest

required_methods = [
    'cluster_list',
    'cluster_get',
    'provision',
    'deprovision',
    'get_kubeconfig',
    'get_progress',
]

engines = [
    JenkinsEngine,
]


class TestBaseEngine:
    @pytest.mark.parametrize('attr_name', ['name', 'verbose_name'])
    def test_class_attrs(self, cluster, attr_name):
        engine = BaseEngine(cluster)

        assert hasattr(engine.__class__, attr_name)

    @pytest.mark.parametrize('attr_name', required_methods)
    def test_object_attrs(self, cluster, attr_name):
        engine = BaseEngine(cluster)
        attr = getattr(engine, attr_name)

        with pytest.raises(NotImplementedError):
            attr()

    def test_engine_status(self, cluster, app):
        engine = BaseEngine(cluster)
        assert engine.engine_status() == app.config.get('PROVISIONER_OK_STATE')

    def test_get_parameter_schema(self, cluster):
        engine = BaseEngine(cluster)
        assert engine.get_parameter_schema() == engine.parameter_schema


class TestAllEngines:
    @pytest.mark.parametrize('engine_class', engines)
    def test_engines_equal_all(self, engine_class):
        assert engine_class.__name__ in all_engines

    @pytest.mark.parametrize('method_name', required_methods)
    @pytest.mark.parametrize('engine_class', engines)
    def test_engine_implements_required_methods(self, engine_class, method_name, cluster):
        engine = engine_class(cluster)

        assert hasattr(engine, method_name)

        method = getattr(engine, method_name)
        try:
            method()
        except NotImplementedError:
            pytest.fail('Engine {} missing method {}'.format(engine_class.__name__, method_name))
        except Exception:
            pass



from kqueen.engines.base import BaseEngine
from kqueen.server import cache
from kqueen.config import current_config

import jenkins
import logging
import requests
import time
import yaml

logger = logging.getLogger(__name__)
config = current_config()


STATE_MAP = {
    'ABORTED': config.get('CLUSTER_ERROR_STATE'),
    'FAILURE': config.get('CLUSTER_ERROR_STATE'),
    'NOT_BUILT': config.get('CLUSTER_UNKNOWN_STATE'),
    'SUCCESS': config.get('CLUSTER_OK_STATE'),
    'UNSTABLE': config.get('CLUSTER_UNKNOWN_STATE')
}


class JenkinsEngine(BaseEngine):
    name = 'jenkins'
    verbose_name = 'Jenkins'
    jenkins_url = config.get('JENKINS_API_URL')
    username = config.get('JENKINS_USERNAME')
    password = config.get('JENKINS_PASSWORD')
    provision_job_name = config.get('JENKINS_PROVISION_JOB_NAME')
    anchor_parameter = config.get('JENKINS_ANCHOR_PARAMETER')
    parameter_schema = {
        'provisioner': {
            'username': {
                'type': 'text',
                'label': 'Username',
                'validators': {
                    'required': True
                }
            },
            'password': {
                'type': 'password',
                'label': 'Password',
                'validators': {
                    'required': True
                }
            }
        },
        'cluster': {}
    }

    def __init__(self, cluster, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.__init__`
        """
        # Call parent init to save cluster on self
        super(JenkinsEngine, self).__init__(cluster, **kwargs)
        # Client initialization
        self.username = kwargs.get('username', self.username)
        self.password = kwargs.get('password', self.password)
        self.client = self._get_client()
        # Cache settings
        self.cache_timeout = 5 * 60

    def _get_provision_job_builds(self):
        """
        Get builds history of Jenkins job used to provision clusters

        Returns:
            dict: More information at :func:`~jenkins.Jenkins.get_job_info`
        """
        return self.client.get_job_info(self.provision_job_name, depth=1)

    @classmethod
    def engine_status(cls):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.engine_status`
        """
        conn_kw = {
            'username': config.get('JENKINS_USERNAME'),
            'password': config.get('JENKINS_PASSWORD')
        }
        status = config.get('PROVISIONER_UNKNOWN_STATE')
        try:
            client = jenkins.Jenkins(config.get('JENKINS_API_URL'), **conn_kw)
            version = client.get_version()
            if version:
                status = config.get('PROVISIONER_OK_STATE')
        except Exception as e:
            logger.error('Could not contact JenkinsEngine backend: {}'.format(repr(e)))
            status = config.get('PROVISIONER_ERROR_STATE')
        return status

    def _get_client(self):
        """
        Initialize Jenkins client

        Returns:
            :obj:`jenkins.Jenkins`: initialized Jenkins client
        """
        return jenkins.Jenkins(self.jenkins_url, **{
            'username': self.username,
            'password': self.password
        })

    def provision(self, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.provision`
        """
        cluster_id = self.cluster.id
        ctx = config.get('JENKINS_PROVISION_JOB_CTX')
        # PATCH THE CTX TO CONTAIN ANCHOR WITH OBJ UUID
        ctx['STACK_NAME'] = 'KQUEEN__{}'.format(cluster_id)
        try:
            self.client.build_job(self.provision_job_name, ctx)
            return True, None
        except Exception as e:
            msg = 'Creating cluster {} failed with following reason: {}'.format(cluster_id, repr(e))
            logger.error(msg)
            return False, msg
        return None, None

    def deprovision(self, **kwargs):
        """
        Deprovisioning isn't supported for Jenkins provisioner yet.

        Implementation of :func:`~kqueen.engines.base.BaseEngine.deprovision`
        """

        return True, None

    def get_kubeconfig(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.get_kubeconfig`
        """
        cluster_external_id = self._get_external_id()
        if not cluster_external_id:
            return {}
        kubeconfig_url = '{jenkins_url}/job/{job_name}/{build_id}/artifact/kubeconfig'.format(
            jenkins_url=self.jenkins_url,
            job_name=self.provision_job_name,
            build_id=str(cluster_external_id),
        )
        kubeconfig = {}
        try:
            kubeconfig = yaml.load(requests.get(kubeconfig_url).text)
        except Exception as e:
            logger.error(repr(e))
        return kubeconfig

    def _get_external_id(self):
        """
        Get external ID of cluster, in this case Jenkins job ID.

        First we try to get external_id from related object metadata, if there is no external_id
        yet, we need to look it up in build history of our configured provisioning Jenkins job

        Returns:
            int: Jenkins job ID
        """
        metadata = self.cluster.metadata or {}
        external_id = metadata.get('external_id', None)
        if external_id:
            return external_id
        try:
            cluster = self._get_by_id()
            external_id = cluster['metadata']['external_id']
            # Get fresh data just in case to avoid conflict
            metadata = self.cluster.metadata or {}
            metadata['external_id'] = external_id
            self.cluster.metadata = metadata
            self.cluster.save()
            return external_id
        except Exception:
            pass
        return external_id

    def _get_by_id(self):
        cluster_id = self.cluster.id
        _list = self.cluster_list()
        cluster = [c for c in _list if c['id'] == cluster_id]
        return cluster[0] if cluster else {}

    def _get_by_external_id(self):
        cluster_external_id = self._get_external_id()
        # Cannot get by external_id if there is no external_id on self.cluster
        if not cluster_external_id:
            return {}
        # Try to get the data from cache
        cluster_cache_key = 'cluster-{}-{}'.format(self.name, cluster_external_id)
        cluster = cache.get(cluster_cache_key)
        if cluster:
            return cluster
        # Get build info for the given job ID (external_id)
        build = self.client.get_build_info(self.provision_job_name, int(cluster_external_id))
        cluster = self._get_cluster_from_build(build)
        return cluster or {}

    def cluster_get(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.cluster_get`

        First we try to get cluster by external_id, because its much more efficient in this
        implementation. If its not possible yet, we return from the slower method
        """
        cluster = self._get_by_external_id()
        if cluster:
            return cluster
        return self._get_by_id()

    def _get_cluster_from_build(self, build):
        cluster_cache_key = 'cluster-{}-{}'.format(self.name, build['number'])
        cluster = cache.get(cluster_cache_key)

        if cluster is None:
            logger.debug('Build {} missing in cache'.format(cluster_cache_key))

            # Prepare build parameters
            _parameters = [d for d in build.get('actions', []) if d.get('parameters', [])]
            parameters = _parameters[0].get('parameters', []) if _parameters else []

            # Try to determine stack name on backend
            stack_name = ''
            if build['result'] in ['SUCCESS'] and build.get('description'):
                stack_name = build['description'].split(' ')[0]

            # Try to determine cluster_id
            _cluster_id = [p.get('value', '') for p in parameters
                           if p.get('name', '') == 'STACK_NAME' and p.get('value', '').startswith('KQUEEN')]
            cluster_id = _cluster_id[0].split('__')[1] if _cluster_id else None

            # Try to determine cluster state
            if build['result']:
                try:
                    state = STATE_MAP[build['result']]
                except KeyError:
                    logger.warning('{} is not valid cluster state'.format(build['result']))
                    state = config.get('CLUSTER_UNKNOWN_STATE')
            else:
                state = config.get('CLUSTER_PROVISIONING_STATE')

            cluster = {
                'key': cluster_cache_key,
                'name': stack_name,
                'id': cluster_id,
                'state': state,
                'metadata': {
                    'external_id': build['number'],
                    'build_timestamp': build['timestamp'],
                    'build_estimated_duration': build['estimatedDuration']
                }
            }

            if cluster['state'] != config.get('CLUSTER_PROVISIONING_STATE'):
                cache.set(cluster_cache_key, cluster, timeout=self.cache_timeout)

        return cluster

    def cluster_list(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.cluster_list`
        """
        job = self._get_provision_job_builds()
        clusters = []

        for build in job['builds']:
            logger.debug('Reading build {}'.format(build))
            cluster = self._get_cluster_from_build(build)
            clusters.append(cluster)

        return clusters

    def get_progress(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.get_progress`
        """
        response = 200
        progress = 1
        result = config.get('CLUSTER_UNKNOWN_STATE')
        try:
            cluster = self.cluster_get()
            result = cluster['state']
            if cluster['state'] == config.get('CLUSTER_PROVISIONING_STATE'):
                # Determine approximate percentage of progress, it is based on estimation
                # from Jenkins, so it can get above 99 percent without being done, so there
                # is patch to hold it on 99 untill its actually done
                now = time.time() * 1000
                start = cluster['metadata']['build_timestamp']
                estimate = cluster['metadata']['build_estimated_duration']
                progress = int(((now - start) / estimate) * 100)
                if progress > 99:
                    progress = 99
            else:
                progress = 100
        except Exception:
            response = 500
        return {'response': response, 'progress': progress, 'result': result}



from kqueen.config import current_config

config = current_config()


class BaseEngine:
    """Base Engine object.

    When you initialize the engine through the prepared property :func:`~kqueen.models.Cluster.engine`
    on :obj:`kqueen.models.Cluster` model object, all keys in engine object parameters attribute (JSONField) on
    :obj:`kqueen.models.Provisioner` object are passed as kwargs.

    Example::
        >>> print my_provisioner.parameters
        {'username': 'foo', 'password': 'bar'}
        >>> print my_cluster.engine.conn_kw
        {'username': 'foo', 'password': 'bar'}

    Credentials passed from parameters attribute to kwargs of MyProvisioner class
    used in conn_kw dict for client initialization.

    Args:
        cluster (:obj:`kqueen.models.Cluster`): Cluster model object related to
        this engine instance.
        **kwargs: Keyword arguments specific to Provisioner implementation.

    Attributes:
        cluster (:obj:`kqueen.models.Cluster`): Cluster model object related to
        this engine instance.
        name (str): Name of the engine usable by program.
        verbose_name (str): Human readable name of the engine.
        parameter_schema (dict): Dictionary representation of the parameters with hints for form rendering.::

            {
                'provisioner': {
                    'username': {
                        'type': 'text',
                        'validators': {
                            'required': True
                        }
                    },
                    'password': {
                        'type': 'password',
                        'validators': {
                            'required': True
                        }
                    }
                }
                'cluster': {
                    'node_count': {
                        'type': 'integer',
                        'validators: {
                            'required': True
                        }
                    }
                }
            }
    """
    name = 'base'
    verbose_name = 'Base Engine'
    parameter_schema = {}

    def __init__(self, cluster, **kwargs):
        self.cluster = cluster

    def cluster_list(self):
        """Get all clusters available on backend.

        Returns:
            list: list of dictionaries. Dictionary format should be::

                {
                    'key': key,     # this record should be cached under this key if you choose to cache
                    'name': name,   # name of the cluster in its respective backend
                    'id': id,       # id of `kqueen.models.Cluster` object in KQueen database
                    'state': state, # cluster.state
                    'metadata': {
                        'foo': bar  # any keys specific for the Provisioner implementation
                    }
                }
        """
        raise NotImplementedError

    def cluster_get(self):
        """Get single cluster from backend related to this engine instance.

        Although this function doesn't take any arguments, it is expected that
        the implementation of the Provisioner gets ``self.cluster`` to provide the
        relevant object for which we want to get data from backend.

        Returns:
            dict: Dictionary format should be::

                {
                    'key': key,     # (str) this record should be cached under this key if you choose to cache
                    'name': name,   # (str) name of the cluster in its respective backend
                    'id': id,       # (str or UUID) id of `kqueen.models.Cluster` object in KQueen database
                    'state': state, # (str) state of cluster on backend represented by app.config['CLUSTER_[FOO]_STATE']
                    'metadata': {
                        'foo': bar  # any keys specific for the Provisioner implementation
                    }
                }
        """
        raise NotImplementedError

    def provision(self):
        """Provision the cluster related to this engine instance to backend.

        Although this function doesn't take any arguments, it is expected that
        the implementation of the Provisioner gets ``self.cluster`` to provide the
        relevant object which we want to provision to backend.

        Returns:
            tuple: First item is bool (success/failure), second item is error, can be None::

                (True, None)                            # successful provisioning
                (False, 'Could not connect to backend') # failed provisioning
        """
        raise NotImplementedError

    def deprovision(self):
        """Deprovision the cluster related to this engine instance from backend.

        Although this function doesn't take any arguments, it is expected that
        the implementation of the Provisioner gets ``self.cluster`` to provide the
        relevant object which we want to remove from backend.

        Returns:
            tuple: First item is bool (success/failure), second item is error, can be None::

                (True, None)                            # successful provisioning
                (False, 'Could not connect to backend') # failed provisioning
        """
        raise NotImplementedError

    def get_kubeconfig(self):
        """Get kubeconfig of the cluster related to this engine from backend.

        Although this function doesn't take any arguments, it is expected that
        the implementation of the Provisioner gets ``self.cluster`` to provide the
        relevant object which we want to get kubeconfig for.

        Returns:
            dict: Dictionary form of kubeconfig (`yaml.load(kubeconfig_file)`)
        """
        raise NotImplementedError

    @classmethod
    def get_parameter_schema(cls):
        """Return parameters specific for this Provisioner implementation.

        This method should return parameters specific to the Provisioner implementation,
        these are used to generate form for creation of Provisioner object and are stored
        in parameters attribute (JSONField) of the `kqueen.models.Provisioner` object.

        Returns:
            dict:  Returns ``self.parameter_schema`` in default, but can be overridden.
        """
        return cls.parameter_schema

    def get_progress(self):
        """Get progress of provisioning if its possible to determine.

        Although this function doesn't take any arguments, it is expected that
        the implementation of the Provisioner gets ``self.cluster`` to provide the
        relevant object which we want to get provisioning progress for.

        Returns:
            dict: Dictionary representation of the provisioning provress.::

                {
                    'response': response, # (int) any number other than 200 means failure to determine progress
                    'progress': progress, # (int) provisioning progress in percents
                    'result': result      # (str) current state of the cluster, i.e. 'Deploying'
                }
        """
        raise NotImplementedError

    @staticmethod
    def engine_status():
        """Check if backend this Provisioner implements is reachable and/or working.

        Returns:
            str: Return status of engine, should use statuses from ``app.config``
        """
        return config.get('PROVISIONER_OK_STATE')



from .aks import AksEngine


class TestAksInit:

    def test_init_aks(self, cluster, monkeypatch):
        def fake_client(self):
            return True

        monkeypatch.setattr(AksEngine, '_get_client', fake_client)
        AksEngine(cluster)



from .manual import ManualEngine
from flask import url_for
from kqueen.conftest import auth_header
from kqueen.conftest import user
from kqueen.models import Cluster
from kqueen.models import Provisioner

import json
import pytest
import yaml

KUBECONFIG = yaml.load(open('kubeconfig_localhost', 'r').read())
CLUSTER_METADATA = {
    'minion_count': 10,
    'cluster_type': 'ha',
    'kubeconfig': KUBECONFIG
}

PROVISIONER_PARAMETERS = {
    'credentials': 'username:password',
}


@pytest.mark.usefixtures('client_class')
class ManualEngineBase:
    def setup(self):
        _user = user()
        create_kwargs_provisioner = {
            'name': 'Testing manual',
            'engine': 'kqueen.engines.ManualEngine',
            'parameters': PROVISIONER_PARAMETERS,
            'owner': _user
        }

        prov = Provisioner(_user.namespace, **create_kwargs_provisioner)
        prov.save(check_status=False)

        self.create_kwargs_cluster = {
            'name': 'Testing cluster for manual provisioner',
            'provisioner': prov,
            'state': 'deployed',
            'kubeconfig': KUBECONFIG,
            'metadata': CLUSTER_METADATA,
            'owner': _user
        }

        self.cluster = Cluster.create(_user.namespace, **self.create_kwargs_cluster)
        self.engine = ManualEngine(cluster=self.cluster)

        # client setup
        self.auth_header = auth_header(self.client)
        self.namespace = self.auth_header['X-Test-Namespace']


class TestClusterAction(ManualEngineBase):
    def test_initialization(self):
        assert self.engine.cluster.get_dict(True) == self.cluster.get_dict(True)

    def test_cluster_list(self):
        assert self.engine.cluster_list() == []

    def test_cluster_get(self):
        assert self.engine.cluster_get() == {}

    @pytest.mark.parametrize('action', ['provision', 'deprovision'])
    def test_actions(self, action):
        method = getattr(self.engine, action)

        assert method() == (True, None)

    def test_get_kubeconfig(self):
        assert self.cluster.engine.get_kubeconfig() == self.cluster.kubeconfig

    def test_progress(self):
        progress = self.engine.get_progress()

        assert 'response' in progress
        assert 'progress' in progress
        assert 'result' in progress

    def test_engine_status(self):
        assert self.engine.engine_status()

    def test_parameter_schema(self):
        assert self.engine.get_parameter_schema() == self.engine.parameter_schema


class TestCreateOverAPI(ManualEngineBase):
    def test_create_over_api(self):
        """Verify Cluster is created over API and kubeconfig is set"""

        url = url_for('api.cluster_list')
        data = self.create_kwargs_cluster
        data['provisioner'] = 'Provisioner:{}'.format(data['provisioner'].id)
        data['owner'] = 'User:{}'.format(data['owner'].id)

        # create
        response = self.client.post(
            url,
            data=json.dumps(data),
            headers=self.auth_header,
            content_type='application/json'
        )

        assert response.status_code == 200

        # load
        cluster_id = response.json['id']
        obj = Cluster.load(self.namespace, cluster_id)
        validation, _ = obj.validate()
        assert validation

        # check parameters
        assert obj.kubeconfig == data['kubeconfig']

        return obj

    @pytest.mark.parametrize('metadata_name', list(CLUSTER_METADATA.keys()))
    def test_metadata_parameters_direct(self, metadata_name):
        assert metadata_name in self.cluster.metadata
        assert self.cluster.metadata[metadata_name] == CLUSTER_METADATA[metadata_name]

    @pytest.mark.parametrize('metadata_name', list(CLUSTER_METADATA.keys()))
    def test_metadata_parameters_api(self, metadata_name):
        obj = self.test_create_over_api()

        assert metadata_name in obj.metadata
        assert obj.metadata[metadata_name] == CLUSTER_METADATA[metadata_name]

    @pytest.mark.parametrize('param_name', list(CLUSTER_METADATA.keys()) + list(PROVISIONER_PARAMETERS.keys()))
    def test_engine_gets_parameters(self, param_name, monkeypatch):
        obj = self.test_create_over_api()

        def fake_init(self, cluster, **kwargs):
            self.cluster = cluster
            self.test_kwargs = kwargs

        monkeypatch.setattr(ManualEngine, '__init__', fake_init)

        engine = obj.engine

        assert param_name in engine.test_kwargs

    def test_delete_run_deprovision(self, monkeypatch):
        """"Deprovision is called before cluster delete"""

        def fake_deprovision(self):
            self.cluster.metadata['_deprovisioned'] = True
            self.cluster.save()

            return True, None

        monkeypatch.setattr(ManualEngine, 'deprovision', fake_deprovision)

        obj = self.test_create_over_api()
        obj.delete()

        assert '_deprovisioned' in obj.metadata
        assert obj.metadata['_deprovisioned']

    def test_delete_remove_object(self):
        obj = self.test_create_over_api()
        obj.delete()

        assert not Cluster.exists(self.namespace, obj.id)



"""Authentication methods for API."""

from kqueen.config import current_config
from kqueen.models import Organization, User
from uuid import uuid4

import bcrypt
import logging

logger = logging.getLogger(__name__)


def authenticate(username, password):
    """
    Authenticate user.

    Args:
        username (str): Username to login
        password (str): Password

    Returns:
        user: authenticated user

    """
    users = list(User.list(None, return_objects=True).values())
    username_table = {u.username: u for u in users}
    user = username_table.get(username)
    if user:
        user_password = user.password.encode('utf-8')
        given_password = password.encode('utf-8')
        if user.active and bcrypt.checkpw(given_password, user_password):
            return user


def identity(payload):
    """
    Read user_id from payload and return User.

    Args:
        payload (dict): Request payload

    Returns:
        user: detected user

    """
    user_id = payload['identity']
    try:
        user = User.load(None, user_id)
    except Exception:
        user = None
    return user


def encrypt_password(_password):
    config = current_config()
    rounds = config.get('BCRYPT_ROUNDS', 12)
    password = str(_password).encode('utf-8')
    encrypted = bcrypt.hashpw(password, bcrypt.gensalt(rounds)).decode('utf-8')
    return encrypted


def is_authorized(_user, policy_value, resource=None):
    """
    Evaluate if given user fulfills requirements of the given
    policy_value.

    Example:
        >>> user.get_dict()
        >>> {'username': 'jsmith', ..., 'role': 'member'}
        >>> is_authorized(user, "ALL")
        True
        >>> is_authorized(user, "IS_ADMIN")
        False

    Args:
        user (dict or User): User data
        policy_value (string or list): Condition written using shorthands and magic keywords

    Returns:
        bool: authorized or not
    """
    if isinstance(_user, User):
        user = _user.get_dict()
    elif isinstance(_user, dict):
        user = _user
    else:
        raise TypeError('Invalid type for argument user {}'.format(type(_user)))

    # magic keywords
    USER = user['id']                       # noqa: F841
    ORGANIZATION = user['organization'].id  # noqa: F841
    ROLE = user['role']
    if resource:
        validation, _ = resource.validate()
        if not validation:
            invalid = True

            # test if we are validating on create view and if so, patch missing object id
            if not resource.id:
                resource.id = uuid4()
                validation, _ = resource.validate()
                if validation:
                    invalid = False
                resource.id = None

            # if invalid resource is passed, let's just continue dispatch_request
            # so it can properly fail with 500 response code
            if invalid:
                logger.error('Cannot evaluate policy for invalid object: {}'.format(str(resource.get_dict())))
                return True

        # TODO: check owner has id and, organization ...
        if hasattr(resource, 'owner'):
            OWNER = resource.owner.id                            # noqa: F841
            OWNER_ORGANIZATION = resource.owner.organization.id  # noqa: F841
        elif isinstance(resource, User):
            OWNER = resource.id                                  # noqa: F841
            OWNER_ORGANIZATION = resource.organization.id        # noqa: F841
        elif isinstance(resource, Organization):
            OWNER_ORGANIZATION = resource.id                     # noqa: F841

    # predefined conditions for evaluation
    conditions = {
        'IS_ADMIN': 'ORGANIZATION == OWNER_ORGANIZATION and ROLE == "admin"',
        'IS_SUPERADMIN': 'ROLE == "superadmin"',
        'IS_OWNER': 'ORGANIZATION == OWNER_ORGANIZATION and USER == OWNER',
        'ADMIN_OR_OWNER': 'ORGANIZATION == OWNER_ORGANIZATION and (ROLE == "admin" or USER == OWNER)',
        'ALL': 'ORGANIZATION == OWNER_ORGANIZATION'
    }

    try:
        condition = conditions[policy_value]
    except KeyError:
        logger.error('Policy evaluation failed. Invalid rule: {}'.format(str(policy_value)))
        return False

    if ROLE == 'superadmin':
        # no point in checking anything here
        return True

    try:
        authorized = eval(condition)
        if not isinstance(authorized, bool):
            logger.error('Policy evaluation did not return boolean: {}'.format(str(authorized)))
            authorized = False
    except Exception as e:
        logger.error('Policy evaluation failed: {}'.format(repr(e)))
        authorized = False
    return authorized



from flask.json import JSONEncoder


class KqueenJSONEncoder(JSONEncoder):
    def default(self, obj):

        if hasattr(obj, 'get_dict'):
            return obj.get_dict()
        elif hasattr(obj, 'serialize'):
            return obj.serialize()

        try:
            return JSONEncoder.default(self, obj)
        except TypeError:
            print('Unserialized')
            print('class', obj.__class__)
            print('bases', dir(obj))
            print(type(obj))

            return {'__{}__'.format(obj.__class__.__name__): obj.__dict__}



from flask import Blueprint
from flask import current_app
from flask import make_response
from flask import request
from flask_jwt import _jwt_required
from ipaddress import ip_address
from ipaddress import ip_network
from prometheus_client import CollectorRegistry
from prometheus_client import CONTENT_TYPE_LATEST
from prometheus_client import generate_latest
from prometheus_client import multiprocess

metrics = Blueprint('metrics', __name__)


@metrics.route('/')
def root():
    # check access
    ip_whitelist = ip_network(current_app.config.get('PROMETHEUS_WHITELIST'))
    if ip_address(request.remote_addr) not in ip_whitelist:
        _jwt_required(current_app.config['JWT_DEFAULT_REALM'])

    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)

    data = generate_latest(registry)

    response = make_response(data)
    response.headers['Content-Type'] = CONTENT_TYPE_LATEST
    response.headers['Content-Length'] = str(len(data))

    return response



"""Kubernetes client wrapper."""


from kubernetes import config, client
from kubernetes.client.rest import ApiException
from kqueen.helpers import prefix_to_num

import logging
import yaml

# define logging
logger = logging.getLogger(__name__)


class KubernetesAPI:
    """Kubernetes API client."""

    def __init__(self, **kwargs):
        """
        Set cluster and prepare clients for all used resource types.

        Args:
            **kwargs: Keyword arguments (cluster is required)
        """
        # load configuration
        try:
            self.cluster = kwargs['cluster']
        except KeyError:
            raise ValueError('Missing parameter cluster')

        logger.debug('Initialized KubernetesAPI for {}'.format(self.cluster))
        self.kubeconfig_file = self.get_kubeconfig_file()

        # set apis
        api_client = config.new_client_from_config(config_file=self.kubeconfig_file)
        self.api_corev1 = client.CoreV1Api(api_client=api_client)
        self.api_extensionsv1beta1 = client.ExtensionsV1beta1Api(api_client=api_client)
        self.api_version = client.VersionApi(api_client=api_client)

    def get_kubeconfig_file(self):
        """
        Write kubeconfig file on filesystem.

        Returns:
            str: Kubeconfig file path

        """
        # TODO: make configfile name random
        configfile = '/tmp/kubernetes'
        f = open(configfile, 'w')
        kubeconfig = self.cluster.get_kubeconfig()
        f.write(yaml.dump(kubeconfig, indent=2))
        f.close()

        return configfile

    def get_version(self):
        """Return Kubernetes version."""
        return self.api_version.get_code().to_dict()

    def list_nodes(self):
        out = []

        try:
            response = self.api_corev1.list_node().items
        except ApiException:
            raise

        for node in response:
            out.append(node.to_dict())

        return out

    def list_persistent_volumes(self):
        out = []

        try:
            response = self.api_corev1.list_persistent_volume().items
        except ApiException:
            raise

        for pv in response:
            out.append(pv.to_dict())

        return out

    def list_persistent_volume_claims(self):
        out = []

        try:
            response = self.api_corev1.list_persistent_volume_claim_for_all_namespaces().items
        except ApiException:
            raise

        for pvc in response:
            out.append(pvc.to_dict())

        return out

    def list_namespaces(self):
        out = []

        try:
            response = self.api_corev1.list_namespace().items
        except ApiException:
            raise

        for namespace in response:
            out.append(namespace.to_dict())

        return out

    def list_pods(self, include_uninitialized=True):
        """List pods in all namespaces."""
        out = []

        try:
            response = self.api_corev1.list_pod_for_all_namespaces(
                include_uninitialized=include_uninitialized
            ).items
        except ApiException:
            raise

        for pod in response:
            out.append(pod.to_dict())

        return out

    def list_pods_by_node(self):
        out = {}
        try:
            nodes = self.list_nodes()
            pods = self.list_pods()
        except ApiException:
            raise

        for node in nodes:
            out[node['metadata']['name']] = []

        for pod in pods:
            node = pod['spec'].get('node_name', 'Unknown')
            out[node].append(pod)

        return out

    def count_pods_by_node(self):
        out = {}

        try:
            pods = self.list_pods_by_node()
        except ApiException:
            raise

        for node_name, pods in pods.items():
            out[node_name] = len(pods)

        return out

    def resources_by_node(self):
        """Read pods on each node, compute sum or requested and limited resources.

        Returns:
            Dict of nodes with allocated resources.
            CPU is float.
            Memory int is in bytes.

        .. code:: yaml

            {
                'node1': {
                    'limits': {'cpu': 2, 'mem': 100},
                    'requests': {'cpu': 1.5, 'mem': 10098}
                }
            }

        """
        out = {}

        try:
            pods = self.list_pods_by_node()
        except ApiException:
            raise

        for node_name, pods in pods.items():
            if node_name not in out:
                out[node_name] = {'limits': {'cpu': 0, 'memory': 0}, 'requests': {'cpu': 0, 'memory': 0}}

            for pod in pods:
                containers = pod.get('spec', {}).get('containers', [])
                for c in containers:
                    resources = c.get('resources')

                    if resources:
                        for resource_policy in ['limits', 'requests']:
                            policy = resources.get(resource_policy, {})

                            if policy:
                                for resource_type in ['cpu', 'memory']:
                                    value = policy.get(resource_type)

                                    if value:
                                        out[node_name][resource_policy][resource_type] += prefix_to_num(value)

        return out

    def _extract_annotation(self, service, prefix='kqueen/'):
        """Read service and return kqueen annotations (if present).

        Args:
            service (dict)
            prefix (str): default `kqueen/`

        Return:
            dict: Annotations matching prefix
        """
        out = {}

        annotations = service.get('metadata', {}).get('annotations', {})

        if annotations:
            for an_name, an in annotations.items():
                if an_name.startswith(prefix):
                    out[an_name[len(prefix):]] = an

        return out

    def list_services(self, include_uninitialized=True, filter_addons=False):
        """List services in all namespaces."""
        out = []

        try:
            response = self.api_corev1.list_service_for_all_namespaces(
                include_uninitialized=include_uninitialized
            ).items
        except ApiException:
            raise

        for item in response:
            if filter_addons:
                addon = self._extract_annotation(item.to_dict())
                if addon:
                    out.append(addon)
            else:
                out.append(item.to_dict())

        return out

    def list_deployments(self, include_uninitialized=True):
        """List deployments in all namespaces."""
        out = []

        try:
            response = self.api_extensionsv1beta1.list_deployment_for_all_namespaces(
                include_uninitialized=include_uninitialized
            ).items
        except ApiException:
            raise

        for item in response:
            out.append(item.to_dict())

        return out

    def list_replica_sets(self, include_uninitialized=True):
        """List replica sets in all namespaces."""
        out = []

        try:
            response = self.api_extensionsv1beta1.list_replica_set_for_all_namespaces(
                include_uninitialized=include_uninitialized
            ).items
        except ApiException:
            raise

        for item in response:
            out.append(item.to_dict())

        return out



"""Configuration and fixtures for pytest."""
from faker import Faker
from kqueen.config import current_config
from kqueen.models import Cluster
from kqueen.models import Organization
from kqueen.models import Provisioner
from kqueen.models import User
from kqueen.server import create_app

import datetime
import etcd
import json
import pytest
import uuid
import yaml

config_file = 'config/test.py'
config = current_config()
fake = Faker()


@pytest.fixture(autouse=True, scope='session')
def app():
    """Prepare app."""
    app = create_app()
    app.testing = True

    return app


@pytest.fixture(autouse=True, scope='session')
def etcd_setup():
    _app = create_app()

    try:
        _app.db.client.delete(_app.config['ETCD_PREFIX'], recursive=True)
    except etcd.EtcdKeyNotFound:
        pass


@pytest.fixture
def cluster():
    """Create cluster with manual provisioner."""
    _uuid = uuid.uuid4()
    _user = user()

    prov = Provisioner(
        _user.namespace,
        name='Fixtured provisioner',
        engine='kqueen.engines.ManualEngine',
        owner=_user
    )
    prov.save(check_status=False)

    create_kwargs = {
        'id': _uuid,
        'name': 'Name for cluster {}'.format(_uuid),
        'provisioner': prov,
        'state': 'deployed',
        'kubeconfig': yaml.load(open('kubeconfig_localhost', 'r').read()),
        'created_at': datetime.datetime(2017, 11, 15, 13, 36, 24),
        'owner': _user
    }

    return Cluster.create(_user.namespace, **create_kwargs)


@pytest.fixture
def provisioner():
    """Create dummy manual provisioner."""
    _user = user()

    create_kwargs = {
        'name': 'Fixtured provisioner',
        'engine': 'kqueen.engines.ManualEngine',
        'owner': _user
    }

    return Provisioner.create(_user.namespace, **create_kwargs)


def get_auth_token(_client, _user):
    """
    Acquire token for given user

    Args:
        client: Client connection
        user: User object

    Returns:
        str: Auth token.
    """

    data = {
        'username': _user.username,
        'password': _user.username + 'password'
    }

    response = _client.post(
        '/api/v1/auth',
        data=json.dumps(data),
        content_type='application/json')

    try:
        token = response.json['access_token']
    except KeyError as e:
        raise KeyError('Unable to read access token from response: {}'.format(response.data))

    return token


@pytest.fixture
def auth_header(client):
    """
    Get JWT access token and convert it to HTTP header.

    Args:
        client: Flask client

    Returns:
        dict: {'Authorization': 'JWT access_token'}

    """
    _user = user()
    token = get_auth_token(client, _user)

    return {
        'Authorization': '{token_prefix} {token}'.format(
            token_prefix=config.get('JWT_AUTH_HEADER_PREFIX'),
            token=token,
        ),
        'X-Test-Namespace': _user.namespace,
        'X-User': str(_user.id),
    }


@pytest.fixture
def organization():
    """Prepare organization object."""
    organization = Organization(
        None,
        name='DemoOrg',
        namespace='demoorg',
    )
    organization.save()

    return organization


@pytest.fixture(scope='class')
def user():
    """Prepare user object."""
    profile = fake.simple_profile()
    user = User.create(
        None,
        username=profile['username'],
        password=profile['username'] + 'password',
        organization=organization(),
        role='superadmin',
        active=True
    )
    user.save()

    return user


@pytest.fixture
def user_with_namespace():

    org = organization()
    org.namespace = fake.user_name()
    org.save()

    _user = user()
    _user.organization = org
    _user.save()

    return _user



from .etcd import EtcdBackend
import pytest


class TestEtcdClient:
    @pytest.fixture(autouse=True)
    def prepare(self, monkeypatch):
        self.set_vars = {
            'KQUEEN_ETCD_HOST': 'etcd-server',
            'KQUEEN_ETCD_PORT': '1234',
        }

        for name, value in self.set_vars.items():
            monkeypatch.setenv(name, value)

    def test_config_fields(self):
        orm = EtcdBackend()
        assert orm.client.host == self.set_vars['KQUEEN_ETCD_HOST']
        assert orm.client.port == int(self.set_vars['KQUEEN_ETCD_PORT'])



from .base import BaseEngine
from kqueen.config import current_config

config = current_config()


class ManualEngine(BaseEngine):
    """
    Manual engine is used for importing existing clusters.
    """

    name = 'manual'
    verbose_name = 'Manual Engine'
    parameter_schema = {
        'provisioner': {},
        'cluster': {
            'kubeconfig': {
                'type': 'yaml_file',
                'label': 'Kubeconfig',
                'validators': {
                    'required': True
                }
            }
        }
    }

    def __init__(self, cluster, **kwargs):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.__init__`
        """
        self.kubeconfig = kwargs.get('kubeconfig', {})
        super(ManualEngine, self).__init__(cluster, **kwargs)

    def cluster_list(self):
        """Manual engine don't support list of clusters"""

        return []

    def cluster_get(self):
        """
        Implementation of :func:`~kqueen.engines.base.BaseEngine.cluster_get`
        """

        return {}

    def provision(self):
        """
        There is no provisioning because Cluster should be already provisioned manually.

        State is updated to OK during in provision method.

        Implementation of :func:`~kqueen.engines.base.BaseEngine.provision`
        """

        self.cluster.state = config.get('CLUSTER_OK_STATE')
        self.cluster.save()

        return True, None

    def deprovision(self):
        """
        Deprovision isn't supported by manual engine, we just pass it.

        Implementation of :func:`~kqueen.engines.base.BaseEngine.deprovision`
        """

        return True, None

    def get_kubeconfig(self):
        """Get kubeconfig of the cluster

        Manual engine don't support any loading of `kubeconfig` so we only return kubeconfig of
        cluster provided during initialization.

        Implementation of :func:`~kqueen.engines.base.BaseEngine.get_kubeconfig`
        """

        return self.kubeconfig

    def get_progress(self):
        """
        Manual engine don't report any progress because cluster is already provisioned before
        cluster is imported

        Implementation of :func:`~kqueen.engines.base.BaseEngine.get_progress`
        """

        return {
            'response': 0,
            'progress': 100,
            'result': config.get('CLUSTER_OK_STATE'),
        }

    @staticmethod
    def engine_status():
        """Manual engine is always available.

        Implementation of :func:`~kqueen.engines.base.BaseEngine.engine_status`
        """
        return config.get('PROVISIONER_OK_STATE')



from kqueen.storages.etcd import BoolField
from kqueen.storages.etcd import DatetimeField
from kqueen.storages.etcd import Field
from kqueen.storages.etcd import IdField
from kqueen.storages.etcd import JSONField
from kqueen.storages.etcd import Model
from kqueen.storages.etcd import ModelMeta
from kqueen.storages.etcd import PasswordField
from kqueen.storages.etcd import RelationField
from kqueen.storages.etcd import StringField
from kqueen.storages.exceptions import BackendError

import datetime
import pytest
import itertools


def create_model(required=False, global_ns=False, encrypted=False):
    class TestModel(Model, metaclass=ModelMeta):
        if global_ns:
            global_namespace = global_ns

        id = IdField(required=required, encrypted=encrypted)
        string = StringField(required=required, encrypted=encrypted)
        json = JSONField(required=required, encrypted=encrypted)
        password = PasswordField(required=required, encrypted=encrypted)
        relation = RelationField(required=required, encrypted=encrypted)
        datetime = DatetimeField(required=required, encrypted=encrypted)
        boolean = BoolField(required=required, encrypted=encrypted)

        _required = required
        _global_ns = global_ns
        _encrypted = encrypted

        if _global_ns:
            _namespace = None
        else:
            _namespace = namespace

    print('Creating model with: required: {}, global_ns: {}, encrypted: {}'.format(
        required,
        global_ns,
        encrypted
    ))

    return TestModel


utcnow = datetime.datetime(1989, 11, 17)

model_kwargs = {'string': 'abc123', 'json': {'a': 1, 'b': 2, 'c': 'tri'}, 'password': 'pass', 'datetime': utcnow, 'boolean': True}
model_kwargs_dict = {'string': 'abc123', 'json': {'a': 1, 'b': 2, 'c': 'tri'}, 'password': 'pass', 'datetime': utcnow.isoformat(), 'boolean': True}
model_fields = ['id', 'string', 'json', 'password', 'relation', 'datetime', 'boolean']
namespace = 'test'


def model_serialized(related=None):
    if related:
        return (
            '{{"string": "abc123", "json": "{{\\"a\\": 1, \\"b\\": 2, \\"c\\": \\"tri\\"}}", '
            '"password": "pass", "relation": "{related_class}:{related_id}", "datetime": {date_timestamp}, '
            '"boolean": "true"}}'.format(
                related_class=related.__class__.__name__,
                related_id=related.id,
                date_timestamp=int(utcnow.timestamp())
            )
        )
    else:
        return (
            '{{"string": "abc123", "json": "{\\"a\\": 1, \\"b\\": 2, \\"c\\": \\"tri\\"}", '
            '"password": "pass", "datetime": {date_timestamp}, "boolean": "true"}}'.format(
                date_timestamp=int(utcnow.timestamp()),
            )
        )


@pytest.fixture(params=itertools.product([True, False], repeat=3))
def get_object(request):
    return create_object(*request.param)


@pytest.fixture(params=itertools.product([True, False], repeat=3))
def get_model(request):
    return create_model(*request.param)


def create_object(required=False, global_ns=False, encrypted=False):
    model = create_model(required, global_ns, encrypted)

    obj1 = model(namespace, **model_kwargs)
    obj2 = model(namespace, **model_kwargs)

    # don't validate first object because we don't have relation field
    obj2.save(False)

    obj1.relation = obj2

    return obj1


class TestModelInit:
    def setup(self):
        self.model = create_model()
        self.obj = self.model(namespace, **model_kwargs)

    @pytest.mark.parametrize('field_name,field_value', model_kwargs.items())
    def test_init_string(self, field_name, field_value):
        """Initialization of new models is properly setting properties"""

        kwargs = {field_name: field_value}
        obj = self.model(namespace, **kwargs)

        assert getattr(obj, field_name) == field_value

    @pytest.mark.parametrize('attr', model_fields)
    @pytest.mark.parametrize('group', ['', '_'])
    def test_field_property_getters(self, attr, group):
        attr_name = '{}{}'.format(group, attr)

        assert hasattr(self.obj, attr_name)


class TestSave:
    def setup(self):
        model = create_model(required=True)
        self.obj = model(namespace)

    def test_model_invalid(self):
        validation, _ = self.obj.validate()

        assert not validation

    def test_save_raises(self):
        with pytest.raises(ValueError, match='Validation for model failed'):
            self.obj.save()

    def test_save_skip_validation(self):
        assert self.obj.save(validate=False)


class TestModelAddId:
    def test_id_added(self, get_object):
        obj = get_object

        assert obj.id is None
        assert obj.verify_id()
        assert obj.id is not None

        obj.save()


class TestRequiredFields:
    @pytest.mark.parametrize('required', [True, False])
    def test_required(self, required):
        model = create_model(required=required)
        obj = model(namespace, **model_kwargs)

        validation, _ = obj.validate()
        assert validation != required


class TestGetFieldNames:
    def test_get_field_names(self, get_object):
        field_names = get_object.__class__.get_field_names()
        req = model_fields

        assert set(field_names) == set(req)

    def test_get_dict(self, get_object):
        dicted = get_object.get_dict()

        assert isinstance(dicted, dict)


class TestFieldSetGet:
    """Validate getters and setters for fields"""
    @pytest.mark.parametrize('field_name', model_kwargs.keys())
    def test_get_fields(self, field_name, get_object):
        at = getattr(get_object, field_name)
        req = model_kwargs[field_name]

        assert at == req

    @pytest.mark.parametrize('field_name', model_kwargs.keys())
    def test_set_fields(self, field_name):
        model_class = create_model()
        obj = model_class(namespace)
        setattr(obj, field_name, model_kwargs[field_name])

        assert getattr(obj, field_name) == model_kwargs[field_name]
        assert obj.get_dict()[field_name] == model_kwargs_dict[field_name]
        assert getattr(obj, '_{}'.format(field_name)).get_value() == model_kwargs[field_name]


class TestSerialization:
    """Serialization and deserialization create same objects"""

    def test_serizalization(self, get_object):
        serialized = get_object.serialize()

        if get_object.__class__._encrypted:
            pytest.skip('Unable to check serialization for encrypted class')

        assert serialized == model_serialized(related=get_object.relation)

    def test_deserialization(self, get_object, monkeypatch):
        def fake(self, class_name):
            return get_object.__class__

        monkeypatch.setattr(RelationField, '_get_related_class', fake)

        object_class = get_object.__class__
        get_object.save()
        new_object = object_class.deserialize(get_object.serialize(), namespace=namespace)

        assert new_object.get_dict(True) == get_object.get_dict(True)


class TestGetDict:
    """Verify objects are serialized properly"""

    def setup(self):
        self.obj1 = create_object()
        self.obj1.string = 'obj1'
        self.obj2 = create_object()
        self.obj1.string = 'obj2'

        self.obj1.save()
        self.obj2.save()

        self.obj1.relation = self.obj2

    def test_get_dict(self):
        dicted = self.obj1.get_dict(expand=False)

        assert dicted['relation'] == self.obj2

    def test_get_dict_expand(self):
        dicted = self.obj1.get_dict(expand=True)

        assert isinstance(dicted['relation'], dict)
        assert dicted['relation'] == self.obj2.get_dict(expand=True)


class TestDuplicateId:
    def setup(self):
        self.model = create_model()

        self.obj1_kwargs = {'string': 'object 1', 'json': {'a': 1, 'b': 2, 'c': 'tri'}, 'password': 'pass'}
        self.obj2_kwargs = {'string': 'object 2', 'json': {'a': 1, 'b': 2, 'c': 'tri'}, 'password': 'pass'}

    def test_with_save(self):
        """"Save object are not same"""
        obj1 = self.model(namespace, **self.obj1_kwargs)
        obj2 = self.model(namespace, **self.obj2_kwargs)

        assert obj1 != obj2

        obj1.save()
        obj2.save()

        print(obj1.get_dict())
        print(obj2.get_dict())

        assert obj1.id != obj2.id


class TestFieldSetValues:
    def setup(self):
        self.value = 'abc123'

    def test_create_by_args(self):
        field = Field(self.value)

        assert field.value == self.value

    def test_create_by_kwargs(self):
        field = Field(value=self.value)

        assert field.value == self.value

    def test_create_by_both(self):
        field = Field('args_value', value='kwargs_value')

        assert field.value == 'args_value'

#
# Relation Field
#


class TestRelationField:
    def setup(self):
        self.obj1 = create_object()
        self.obj1.name = 'obj1'
        self.obj2 = create_object()
        self.obj2.name = 'obj2'

        self.obj1.save()
        self.obj2.save()

        self.obj1.relation = self.obj2

    def test_relation_attrs(self):
        assert self.obj1.relation == self.obj2
        assert self.obj1._relation.value == self.obj2

    def test_relation_serialization(self):
        ser = self.obj1._relation.serialize()
        req = '{model_name}:{object_id}'.format(
            model_name=self.obj2.__class__.__name__,
            object_id=self.obj2.id
        )

        assert ser == req

    def test_relation_loading(self, monkeypatch):
        def fake_related_class(their, class_name):
            return self.obj1.__class__

        monkeypatch.setattr(RelationField, '_get_related_class', fake_related_class)

        self.obj1.save()
        loaded = self.obj1.__class__.load(namespace, self.obj1.id)

        assert isinstance(loaded, self.obj1.__class__)
        assert hasattr(loaded, 'relation')
        assert loaded.relation == self.obj2


class TestNamespaces:
    def setup(self):
        self.class_namespaced = create_model(False, False)
        self.class_global = create_model(False, True)

    def test_is_namespaced(self):
        assert self.class_namespaced.is_namespaced()
        assert not self.class_global.is_namespaced()

    def test_write_with_namespace(self):
        obj = self.class_namespaced(namespace, **model_kwargs)

        assert obj.save()
        assert obj._object_namespace == namespace

    def test_raise_missing_namespace(self):
        with pytest.raises(BackendError):
            self.class_namespaced(**model_kwargs)

    @pytest.mark.parametrize('ns,req_ns', [
        ('test', 'test')
    ])
    def test_get_db_prefix(self, ns, req_ns):
        cls = self.class_namespaced
        prefix = cls.get_db_prefix(ns)

        req = '/{app_prefix}/obj/{namespace}/{class_name}/'.format(
            app_prefix='kqueen_test',
            namespace=req_ns,
            class_name=cls.get_model_name(),
        )

        assert req == prefix

    def test_get_db_prefix_raises_on_missing_ns(self):
        cls = self.class_namespaced

        with pytest.raises(BackendError, match='Missing namespace'):
            cls.get_db_prefix(None)


datetime_sample = datetime.datetime(2007, 12, 6, 16, 29, 43)


class TestDateTimeField:
    def setup(self):
        self.datetime = datetime_sample
        self.field = DatetimeField()

    def test_serialization(self):
        req = int(self.datetime.timestamp())
        self.field.set_value(self.datetime)

        assert self.field.serialize() == req

    def test_serialization_none(self):
        self.field.set_value(None)

        assert self.field.serialize() is None

    @pytest.mark.parametrize('serialized, req', [
        (datetime_sample.timestamp(), datetime_sample),
        (int(datetime_sample.timestamp()), datetime_sample),
        (datetime_sample.isoformat(), datetime_sample),
    ])
    def test_deserialization(self, serialized, req):
        self.field.deserialize(serialized)

        assert self.field.value == req

    def test_dict_value_returns_isoformat(self):
        self.field.set_value(self.datetime)

        assert self.field.dict_value() == self.field.value.isoformat()


class TestBoolField:
    def setup(self):
        self.boolean = False
        self.field = BoolField()

    def test_serialization(self):
        req = 'false'
        self.field.set_value(self.boolean)

        assert self.field.serialize() == req

    def test_serialization_none(self):
        self.field.set_value(None)

        assert self.field.serialize() is None

    @pytest.mark.parametrize('serialized, req', [
        ('false', False),
        ('true', True),
    ])
    def test_deserialization(self, serialized, req):
        self.field.deserialize(serialized)

        assert self.field.value == req

    def test_dict_value_returns_boolean(self):
        self.field.set_value(self.boolean)

        assert self.field.dict_value() == self.boolean


#
# Encryption
#
class TestFieldEncryption:
    def test_get_encryption_key(self, get_model):

        obj = get_model(get_model._namespace, **model_kwargs)

        field = obj._string
        KEY_LENGTH = obj._string.bs
        key = field._get_encryption_key()

        assert len(key) == KEY_LENGTH

    @pytest.mark.parametrize('field_name', model_kwargs.keys())
    def test_encrypt_none(self, field_name):
        field_value = None

        cls = create_model(False, False, True)
        obj = cls(namespace, **model_kwargs)

        field = getattr(obj, '_{}'.format(field_name))
        field.value = field_value
        assert field.encrypted
        assert field.encrypt() is None

    @pytest.mark.parametrize('field_value', [i * 'a' for i in range(35)])
    def test_string_various_length(self, field_value):
        field_name = 'string'
        cls = create_model(False, False, True)
        obj = cls(namespace, **model_kwargs)

        field = getattr(obj, '_{}'.format(field_name))
        field.value = field_value
        assert field.encrypted

        encrypted = field.encrypt()

        field.value = None
        field.decrypt(encrypted)

        assert field.value == field_value

    @pytest.mark.parametrize('field_name, field_value', model_kwargs.items())
    def test_serialized_model_dont_contain_value(self, field_name, field_value):
        cls = create_model(False, False, True)
        obj = cls(namespace, **model_kwargs)

        serialized = obj.serialize()
        field = getattr(obj, '_{}'.format(field_name))

        assert '"{}"'.format(field.value) not in serialized
        assert '"{}"'.format(field.serialize()) not in serialized


class TestModelEncryptionWithNone:
    def test_serialization(self, monkeypatch, get_object):
        def fake(self, class_name):
            return get_object.__class__

        monkeypatch.setattr(RelationField, '_get_related_class', fake)

        # load information about test setup
        namespace = get_object.__class__._namespace
        required = get_object.__class__._required

        # set None to fields if possible
        if not required:
            for field_name in get_object.__class__.get_fields().keys():
                field = getattr(get_object, '_{}'.format(field_name))
                field.value = None

                assert field.value is None

        get_object.save()

        loaded = get_object.__class__.load(namespace, get_object.id)

        assert get_object.get_dict(True) == loaded.get_dict(True)



from .helpers import get_object
from kqueen.conftest import cluster
from kqueen.conftest import user
from kqueen.storages.exceptions import BackendError

import pytest


class TestGetObject:
    def setup(self):
        self.obj = cluster()
        self.obj.save()
        self.user = user()

    def test_get_objects(self):
        obj = get_object(self.obj.__class__, self.obj.id, self.user)

        assert obj.get_dict(True) == self.obj.get_dict(True)

    @pytest.mark.parametrize('bad_user', [
        'None',
        '',
        None,
        {},
    ])
    def test_get_object_malformed_user(self, bad_user):
        with pytest.raises(BackendError, match='Missing namespace for class'):
            get_object(self.obj.__class__, self.obj.id, bad_user)



from .test_crud import BaseTestCRUD
from flask import url_for
from kqueen.conftest import provisioner
from kqueen.engines.__init__ import __all__ as all_engines
from pprint import pprint as print


class TestProvisionerCRUD(BaseTestCRUD):
    def get_object(self):
        return provisioner()

    def get_edit_data(self):
        return {
            'name': 'patched cluster',
            'engine': 'kqueen.engines.Dummy',
        }

    def get_create_data(self):
        data = self.obj.get_dict()
        data['id'] = None
        data['owner'] = 'User:{}'.format(self.obj.owner.id)

        return data

    def test_provisioner_engines(self):
        url = url_for('api.provisioner_engine_list')

        response = self.client.get(url, headers=self.auth_header)
        print(response.json)

        assert response.status_code == 200
        assert isinstance(response.json, list)
        assert len(response.json) == len(all_engines)



from .generic_views import CreateView
from .generic_views import DeleteView
from .generic_views import GetView
from .generic_views import ListView
from .generic_views import UpdateView
from .helpers import get_object
from flask import abort
from flask import Blueprint
from flask import jsonify
from flask import make_response
from flask import request
from flask_jwt import current_identity
from flask_jwt import jwt_required
from importlib import import_module
from kqueen.auth import encrypt_password
from kqueen.models import Cluster
from kqueen.models import Organization
from kqueen.models import Provisioner
from kqueen.models import User
from kqueen.config import current_config

import asyncio
import logging
import os
import yaml

config = current_config()
logger = logging.getLogger(__name__)

api = Blueprint('api', __name__)


# error handlers
def error_response(code, error):
    """Return JSONed response of error code.

    Attributes:
        code (int): Error code number.
        error (obj): HTTP error code

    Returns:
        JSONified error response
    """

    response = {'code': code, 'description': error.description}
    return make_response(jsonify(response), code)


@api.errorhandler(400)
def bad_request(error):
    return error_response(400, error)


@api.errorhandler(404)
def not_found(error):
    return error_response(404, error)


@api.errorhandler(500)
def not_implemented(error):
    return error_response(500, error)


@api.route('/')
@api.route('/health')
def index():
    return jsonify({'response': 'Gutten tag!'})


# Clusters
class ListClusters(ListView):
    object_class = Cluster

    async def _update_clusters(self, clusters):
        loop = asyncio.get_event_loop()
        futures = [loop.run_in_executor(None, c.get_state) for c in clusters]

        for _ in await asyncio.gather(*futures):
            pass

    def get_content(self, *args, **kwargs):
        clusters = super(ListClusters, self).get_content(*args, **kwargs)

        if config.get('CLUSTER_STATE_ON_LIST'):
            try:
                loop = asyncio.get_event_loop()
                loop.run_until_complete(self._update_clusters(clusters))
            except RuntimeError:
                logger.warning('Asyncio loop is NOT available, fallback to simple looping')

                for c in clusters:
                    c.get_state()

        return clusters


class CreateCluster(CreateView):
    object_class = Cluster

    def after_save(self):
        # start provisioning
        prov_status, prov_msg = self.obj.engine.provision()

        if not prov_status:
            logger.error('Provisioning failed: {}'.format(prov_msg))
            abort(500, description=prov_msg)


class GetCluster(GetView):
    object_class = Cluster

    def dispatch_request(self, *args, **kwargs):
        self.check_authentication()
        self.set_object(*args, **kwargs)
        self.check_authorization()
        cluster = self.get_content(*args, **kwargs)
        cluster.get_state()

        return jsonify(cluster)


class UpdateCluster(UpdateView):
    object_class = Cluster


class DeleteCluster(DeleteView):
    object_class = Cluster


api.add_url_rule('/clusters', view_func=ListClusters.as_view('cluster_list'))
api.add_url_rule('/clusters', view_func=CreateCluster.as_view('cluster_create'))
api.add_url_rule('/clusters/<uuid:pk>', view_func=GetCluster.as_view('cluster_get'))
api.add_url_rule('/clusters/<uuid:pk>', view_func=UpdateCluster.as_view('cluster_update'))
api.add_url_rule('/clusters/<uuid:pk>', view_func=DeleteCluster.as_view('cluster_delete'))


@api.route('/clusters/<uuid:pk>/status', methods=['GET'])
@jwt_required()
def cluster_status(pk):
    obj = get_object(Cluster, pk, current_identity)

    return jsonify(obj.status())


@api.route('/clusters/<uuid:pk>/topology-data', methods=['GET'])
@jwt_required()
def cluster_topology_data(pk):
    obj = get_object(Cluster, pk, current_identity)

    return jsonify(obj.topology_data())


@api.route('/clusters/<uuid:pk>/kubeconfig', methods=['GET'])
@jwt_required()
def cluster_kubeconfig(pk):
    obj = get_object(Cluster, pk, current_identity)

    return jsonify(obj.kubeconfig)


@api.route('/clusters/<uuid:pk>/progress', methods=['GET'])
@jwt_required()
def cluster_progress(pk):
    obj = get_object(Cluster, pk, current_identity)
    try:
        progress = obj.engine.get_progress()
    except NotImplementedError:
        progress = {
            'response': 501,
            'progress': 0,
            'result': obj.get_state()
        }
    except Exception:
        progress = {
            'response': 500,
            'progress': 0,
            'result': config.get('CLUSTER_UNKNOWN_STATE')
        }
    return jsonify(progress)


# Provisioners
class ListProvisioners(ListView):
    object_class = Provisioner


class CreateProvisioner(CreateView):
    object_class = Provisioner


class GetProvisioner(GetView):
    object_class = Provisioner


class UpdateProvisioner(UpdateView):
    object_class = Provisioner


class DeleteProvisioner(DeleteView):
    object_class = Provisioner


api.add_url_rule('/provisioners', view_func=ListProvisioners.as_view('provisioner_list'))
api.add_url_rule('/provisioners', view_func=CreateProvisioner.as_view('provisioner_create'))
api.add_url_rule('/provisioners/<uuid:pk>', view_func=GetProvisioner.as_view('provisioner_get'))
api.add_url_rule('/provisioners/<uuid:pk>', view_func=UpdateProvisioner.as_view('provisioner_update'))
api.add_url_rule('/provisioners/<uuid:pk>', view_func=DeleteProvisioner.as_view('provisioner_delete'))


@api.route('/provisioners/engines', methods=['GET'])
@jwt_required()
def provisioner_engine_list():
    engine_cls = []
    module_path = 'kqueen.engines'

    for engine in Provisioner.list_engines():
        try:
            module = import_module(module_path)
            _class = getattr(module, engine)
            parameters = _class.get_parameter_schema()
            name = '.'.join([module_path, engine])
            engine_cls.append({
                'name': name,
                'verbose_name': getattr(_class, 'verbose_name', name),
                'parameters': parameters
            })
        except NotImplementedError:
            engine_cls.append({
                'name': engine,
                'verbose_name': engine,
                'parameters': {
                    'provisioner': {},
                    'cluster': {}
                }
            })
        except Exception:
            logger.error('Unable to read parameters for engine {}'.format(engine))

    return jsonify(engine_cls)


# Organizations
class ListOrganizations(ListView):
    object_class = Organization


class CreateOrganization(CreateView):
    object_class = Organization


class GetOrganization(GetView):
    object_class = Organization


class UpdateOrganization(UpdateView):
    object_class = Organization


class DeleteOrganization(DeleteView):
    object_class = Organization


api.add_url_rule('/organizations', view_func=ListOrganizations.as_view('organization_list'))
api.add_url_rule('/organizations', view_func=CreateOrganization.as_view('organization_create'))
api.add_url_rule('/organizations/<uuid:pk>', view_func=GetOrganization.as_view('organization_get'))
api.add_url_rule('/organizations/<uuid:pk>', view_func=UpdateOrganization.as_view('organization_update'))
api.add_url_rule('/organizations/<uuid:pk>', view_func=DeleteOrganization.as_view('organization_delete'))


@api.route('/organizations/<uuid:pk>/policy', methods=['GET'])
@jwt_required()
def organization_policy(pk):
    obj = get_object(Organization, pk, current_identity)
    policies = config.get('DEFAULT_POLICIES', {})
    if hasattr(obj, 'policy') and obj.policy:
        policies.update(obj.policy)

    return jsonify(policies)


# Users
class ListUsers(ListView):
    object_class = User


class CreateUser(CreateView):
    object_class = User


class GetUser(GetView):
    object_class = User


class UpdateUser(UpdateView):
    object_class = User

    def dispatch_request(self, *args, **kwargs):
        self.check_authentication()

        if not request.json:
            abort(400, description='JSON data expected')

        data = request.json
        if not isinstance(data, dict):
            abort(400)

        self.set_object(*args, **kwargs)

        if 'password' in data:
            del data['password']

        for key, value in data.items():
            setattr(self.obj, key, value)

        try:
            self.obj.save()
        except Exception:
            abort(500)

        output = self.get_content(*args, **kwargs)
        return jsonify(output)


class DeleteUser(DeleteView):
    object_class = User


api.add_url_rule('/users', view_func=ListUsers.as_view('user_list'))
api.add_url_rule('/users', view_func=CreateUser.as_view('user_create'))
api.add_url_rule('/users/<uuid:pk>', view_func=GetUser.as_view('user_get'))
api.add_url_rule('/users/<uuid:pk>', view_func=UpdateUser.as_view('user_update'))
api.add_url_rule('/users/<uuid:pk>', view_func=DeleteUser.as_view('user_delete'))


@api.route('/users/<uuid:pk>/updatepw', methods=['PATCH'])
@jwt_required()
def user_password_update(pk):
    obj = get_object(User, pk, current_identity)

    if not request.json:
        abort(400, description='JSON data expected')

    data = request.json
    if not isinstance(data, dict):
        abort(400)

    obj.password = encrypt_password(data.get('password'))

    try:
        obj.save()
    except Exception:
        abort(500)

    return jsonify(obj)


@api.route('/users/whoami', methods=['GET'])
@jwt_required()
def user_whoami():
    output = current_identity

    return jsonify(output)


@api.route('/swagger', methods=['GET'])
def swagger_json():
    try:
        base_path = os.path.abspath(os.path.dirname(__file__))
        file_path = os.path.join(base_path, 'api.yml')
        with open(file_path, 'r') as f:
            _yaml = f.read()
        data = yaml.safe_load(_yaml)
    except FileNotFoundError:
        logger.error('Swagger YAML not found on {}.'.format(file_path))
        abort(404)
    except Exception as e:
        logger.error(e)
        abort(500)

    return jsonify(data)



from .exceptions import BackendError
from Crypto import Random
from Crypto.Cipher import AES
from datetime import datetime
from dateutil.parser import parse as du_parse
from flask import current_app
from kqueen.config import current_config

import base64
import etcd
import hashlib
import importlib
import json
import logging
import uuid

logger = logging.getLogger(__name__)


class EtcdBackend:
    def __init__(self, **kwargs):
        config = current_config()

        self.client = etcd.Client(
            host=config.get('ETCD_HOST', 'localhost'),
            port=int(config.get('ETCD_PORT', 4001)),
        )
        self.prefix = '{}/obj/'.format(config.get('ETCD_PREFIX', '/kqueen'))


class Field:
    is_field = True

    def __init__(self, *args, **kwargs):
        """Initialize Field object

        Argss:
            value: Set field value. This has higher priority than using attributes.

        Attributes:
            required (bool): Set field to be required before saving the model. Defaults to False.
            value: Set field value.

        """

        # value can be passed as args[0] or kwargs['value']
        if len(args) >= 1:
            self.value = args[0]
        else:
            self.value = kwargs.get('value', None)

        # set block size for crypto
        self.bs = 16

        # field parameters
        self.required = kwargs.get('required', False)
        self.encrypted = kwargs.get('encrypted', False)

    def on_create(self, **kwargs):
        """Optional action that should be run only on newly created objects"""
        pass

    def set_value(self, value, **kwargs):
        self.value = value

    def get_value(self):
        return self.value

    def dict_value(self):
        """Return field representation for API"""

        return self.get_value()

    def serialize(self):

        if self.value is not None:
            return str(self.value)
        else:
            return None

    def deserialize(self, serialized, **kwargs):
        """
        This method is used for value deserialization. It is necessary to create instance first
        (with empty value) and then use `deserialize` method to fill the value.

        Default implementation in `Field` class is only passing value. It should be extended in
        specific field classes.

        Attributes:
            serialized (string): Serialized value of the field.
        """

        self.set_value(serialized, **kwargs)

    def empty(self):
        return self.value is None

    def validate(self):
        """
        This method is called before saving model and can be used to validate format or
        consitence of fields

        Returns:
            Result of validation. True for success, False otherwise.
        """
        return True

    def _get_encryption_key(self):
        """
        Read encryption key and format it.

        Returns:
            Encryption key.
        """

        # check for key
        config = current_config()
        key = config.get('SECRET_KEY')

        if key is None:
            raise Exception('Missing SECRET_KEY')

        # calculate hash passowrd
        return hashlib.sha256(key.encode('utf-8')).digest()[:self.bs]

    def _pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    def _unpad(self, s):
        return s[:-ord(s[len(s) - 1:])]

    def encrypt(self):
        """Encrypt stored value"""

        serialized = self.serialize()

        if not self.encrypted:
            return serialized

        if serialized is not None:
            key = self._get_encryption_key()
            padded = self._pad(str(serialized))

            iv = Random.new().read(self.bs)
            suite = AES.new(key, AES.MODE_CBC, iv)
            encrypted = suite.encrypt(padded)
            encoded = base64.b64encode(iv + encrypted).decode('utf-8')

            return encoded

    def decrypt(self, crypted, **kwargs):

        if not self.encrypted:
            return self.deserialize(crypted, **kwargs)

        key = self._get_encryption_key()
        decoded = base64.b64decode(crypted)

        iv = decoded[:self.bs]
        suite = AES.new(key, AES.MODE_CBC, iv)
        decrypted = suite.decrypt(decoded[self.bs:])
        decrypted_decoded = decrypted.decode('utf-8')

        serialized = self._unpad(decrypted_decoded)

        self.deserialize(serialized, **kwargs)

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        # second is field
        if hasattr(other, 'is_field') and other.is_field:
            return self.value == other.value
        else:
            return self.value == other


class StringField(Field):
    pass


class BoolField(Field):

    def deserialize(self, serialized, **kwargs):
        if isinstance(serialized, str):
            value = json.loads(serialized)
            self.set_value(value, **kwargs)

    def set_value(self, value, **kwargs):
        if value is not None:
            if isinstance(value, bool):
                self.value = value
            else:
                self.deserialize(value)

    def serialize(self):
        if isinstance(self.value, bool):
            return json.dumps(self.value)


class IdField(Field):
    def set_value(self, value, **kwargs):
        """Don't serialize None"""
        if value is not None:
            self.value = str(value)
        else:
            self.value = value


class PasswordField(Field):

    def on_create(self):
        from kqueen.auth import encrypt_password
        self.value = encrypt_password(self.value)


class DatetimeField(Field):
    """
    Datetime is stored as UTC timestamp in DB and as naive datetime on instance.
    """

    def deserialize(self, serialized, **kwargs):
        value = None

        # convert to float if serialized is digit
        if isinstance(serialized, str) and serialized.isdigit():
            serialized = float(serialized)

        if isinstance(serialized, (float, int)):
            value = datetime.fromtimestamp(serialized)
            self.set_value(value, **kwargs)
        elif isinstance(serialized, str):
            value = du_parse(serialized)
            self.set_value(value, **kwargs)

    def set_value(self, value, **kwargs):
        if value and isinstance(value, datetime):
            self.value = value
        else:
            self.deserialize(value)

    def serialize(self):
        if isinstance(self.value, datetime):
            return int(self.value.timestamp())
        else:
            return None

    def dict_value(self):
        """Return API representation of value"""

        if self.value and isinstance(self.value, datetime):
            return self.value.isoformat()
        else:
            return None


class JSONField(Field):
    """JSON is stored as value"""

    def set_value(self, value, **kwargs):
        if isinstance(value, str):
            self.value = json.loads(value)
        elif isinstance(value, dict):
            self.value = value

    def serialize(self):
        if isinstance(self.value, dict):
            return json.dumps(self.value)
        else:
            return None


class RelationField(Field):
    """Store relations between models.

    Serialization format is `ModelName:object_id`.

    """

    # TODO: make Model property - limit relation objects only to one model

    def serialize(self):
        if self.value and self.__class__.is_field:
            return '{model_name}:{object_id}'.format(
                model_name=self.value.__class__.__name__,
                object_id=self.value.id,
            )
        else:
            return None

    def deserialize(self, serialized, **kwargs):
        """Deserialize relation to real object"""

        if ':' in serialized:
            class_name, object_id = serialized.split(':')

            obj_class = self._get_related_class(class_name)
            # TODO: CRITICAL make namespaced and check it

            obj = obj_class.load(kwargs.get('namespace'), object_id)
            self.set_value(obj, **kwargs)

    def _get_related_class(self, class_name):
        module = importlib.import_module('kqueen.models')

        return getattr(module, class_name)

    def validate(self):
        try:
            class_name = self.value.__class__.__name__
            selfid = self.value.id
        except Exception:
            return False

        return class_name and selfid

    def set_value(self, value, **kwargs):
        """Detect serialized format and deserialized according to format."""
        if isinstance(value, str) and ':' in value:
            # deserialize
            self.deserialize(value, **kwargs)
        else:
            super(RelationField, self).set_value(value, **kwargs)


class ModelMeta(type):
    def __new__(cls, clsname, superclasses, attributedict):
        newattributes = attributedict.copy()
        fields = {}

        # loop attributes and set getters and setter for Fields
        for attr_name, attr in attributedict.items():
            attr_class = attr.__class__
            if hasattr(attr_class, 'is_field') and attr_class.is_field:
                name_hidden = '_{}'.format(attr_name)
                fields[attr_name] = attr

                def fget(self, k=attr_name):
                    att = getattr(self, "_{}".format(k))
                    return att.get_value()

                def fset(self, value, k=attr_name):
                    att = getattr(self, "_{}".format(k))
                    att.set_value(value)

                newattributes[attr_name] = property(fget, fset)
                logger.debug('Setting {} to point to {}'.format(attr_name, name_hidden))

        newattributes['_fields'] = fields

        return type.__new__(cls, clsname, superclasses, newattributes)


class Model:
    """Parent class for all models"""
    # id field is required for all models
    id = IdField()

    def __init__(self, ns=None, **kwargs):
        """
        Create model object.

        Args:
            ns (str): Namespace for created object. Required for namespaced objects.

        Attributes:
            **kwargs: Object attributes.

        """

        # manage namespace
        if self.__class__.is_namespaced():
            self._object_namespace = ns

            if not self._object_namespace:
                raise BackendError('Missing namespace for class {}'.format(self.__class__.__name__))

        # loop fields and set it
        for field_name, field in self.__class__.get_fields().items():
            field_class = field.__class__
            if hasattr(field_class, 'is_field'):
                field_object = field_class(**field.__dict__)
                field_object.set_value(kwargs.get(field_name), namespace=ns)

                # Hash password field in case of new DB entry
                if kwargs.get('__create__', False):
                    field_object.on_create()

                setattr(self, '_{}'.format(field_name), field_object)

    @classmethod
    def get_model_name(cls):
        """Return lowercased name of the class"""

        return cls.__name__.lower()

    @classmethod
    def is_namespaced(cls):
        """
        Check if model is namespaced or global.

        Returns:
            bool: True for namespaced models, False for global models.
        """
        if hasattr(cls, 'global_namespace') and cls.global_namespace:
            return False
        else:
            return True

    @classmethod
    def get_db_prefix(cls, namespace=None):
        """Calculate prefix for writing DB objects

        Returns:
            string: Database prefix

        Example:

            /kqueen/default/MyModel/

        """

        if cls.is_namespaced():
            if not namespace:
                raise BackendError('Missing namespace for class {}'.format(cls.__name__))
        else:
            namespace = 'global'

        return '{prefix}{namespace}/{model}/'.format(
            prefix=current_app.db.prefix,
            namespace=namespace,
            model=cls.get_model_name(),
        )

    @classmethod
    def create(cls, ns, **kwargs):
        """Create new object"""

        kwargs['__create__'] = True
        o = cls(ns, **kwargs)

        return o

    @classmethod
    def list(cls, namespace, return_objects=True):
        """List objects in the database"""
        output = {}

        key = cls.get_db_prefix(namespace)

        try:
            directory = current_app.db.client.get(key)
        except etcd.EtcdKeyNotFound:
            return output

        # Don't allow iteration over children generator on empty directory
        # More informations here: https://github.com/jplana/python-etcd/issues/54
        if not getattr(directory, '_children', []):
            return output

        for result in directory.children:
            if return_objects:
                output[result.key.replace(key, '')] = cls.deserialize(result.value, namespace=namespace)
            else:
                output[result.key.replace(key, '')] = None

        return output

    @classmethod
    def load(cls, namespace, object_id):
        """Load object from database"""

        key = '{}{}'.format(cls.get_db_prefix(namespace), str(object_id))
        try:
            response = current_app.db.client.read(key)
            value = response.value
        except etcd.EtcdKeyNotFound:
            raise NameError('Object not found')
        except Exception:
            raise

        return cls.deserialize(value, key=key, namespace=namespace)

    @classmethod
    def exists(cls, namespace, object_id):
        """Check if object exists"""

        try:
            cls.load(namespace, object_id)
            return True
        except NameError:
            return False

    @classmethod
    def deserialize(cls, serialized, **kwargs):
        object_kwargs = {}
        # deserialize toplevel dict and loop fields and deserialize them
        toplevel = json.loads(serialized)

        for field_name, field in cls.get_fields().items():
            field_class = field.__class__

            if hasattr(field_class, 'is_field') and toplevel.get(field_name) is not None:
                field_object = field_class(**field.__dict__)
                field_object.decrypt(toplevel[field_name], **kwargs)

                object_kwargs[field_name] = field_object.get_value()

        o = cls(kwargs.get('namespace'), **object_kwargs)

        if kwargs.get('key'):
            o._key = kwargs.get('key')

        return o

    @classmethod
    def get_fields(cls):
        """Return dict of fields and it classes"""

        return cls._fields

    @classmethod
    def get_field_names(cls):
        """Return list of field names"""

        return list(cls._fields.keys())

    def get_db_key(self):
        if not self.id:
            raise Exception('Missing object id')

        if self.__class__.is_namespaced():
            namespace = self._object_namespace
        else:
            namespace = None

        return '{}{}'.format(self.__class__.get_db_prefix(namespace), self.id)

    def verify_id(self):
        if hasattr(self, 'id') and self.id is not None:
            return self.id
        else:
            newid = uuid.uuid4()

            # TODO check id doesn't exists

            self.id = newid
            return self.id

    def save(self, validate=True, assign_id=True):
        """Save object to database


        Attributes:
            validate (bool): Validate model before saving. Defaults to `True`.
            assign_id (bool): Assing id (if missing) before saving model. Defaults to `True`

        Return:
            bool: `True` if model was saved without errors, `False` otherwise.

        """

        if assign_id:
            self.verify_id()

        validation_status, validation_msg = self.validate()
        if validate and not validation_status:
            raise ValueError('Validation for model failed with: {}'.format(validation_msg))

        key = self.get_db_key()
        logger.debug('Writing {} to {}'.format(self, key))

        try:
            current_app.db.client.write(key, self.serialize())

            self._key = key
            return True
        except Exception:
            raise

    def delete(self):
        """Delete the object"""

        current_app.db.client.delete(self.get_db_key())

    def validate(self):
        """Validate the model object pass all requirements

        Checks:
            * Required fields

        Returns:
            Validation result. `True` for passed, `False` for failed.
        """

        fields = self.__class__.get_field_names()
        for field in fields:
            hidden_field = '_{}'.format(field)
            field_object = getattr(self, hidden_field)

            # validation
            # TODO: move to validate method of Field
            if field_object.required and field_object.value is None:
                return False, 'Required field {} is None'.format(field)

            if field_object.value and not field_object.validate():
                return False, 'Field {} validation failed'.format(field)

        return True, None

    def _expand(self, obj):
        expanded = obj.get_dict()
        for key, value in expanded.items():
            if hasattr(value, 'get_dict'):
                expanded[key] = self._expand(value)
        return expanded

    def get_dict(self, expand=False):
        """Return object properties represented by dict.

        Attributes:
            expand (bool): Expand properties to dict (if possible).

        Returns:
            Dict with object properties
        """

        output = {}

        for field_name in self.__class__.get_field_names():
            field = getattr(self, '_{}'.format(field_name))

            if expand and hasattr(field.value, 'get_dict'):
                wr = self._expand(field.value)
            elif hasattr(field, 'dict_value'):
                wr = field.dict_value()
            else:
                wr = field.get_value()

            if wr is not None:
                output[field_name] = wr

        return output

    def serialize(self):
        serdict = {}
        for attr_name, attr in self.get_dict().items():
            value = getattr(self, '_{}'.format(attr_name)).encrypt()
            if value is not None:
                serdict[attr_name] = value

        return json.dumps(serdict)

    def __str__(self):
        return '{} <{}>'.format(self.__class__.get_model_name(), self.id)

    def __eq__(self, other):
        if hasattr(other, 'serialize'):
            return self.serialize() == other.serialize()
        else:
            return False

# TODO: implement autogenerated fields (generete them if missing)
# TODO: implement unique field:w
# TODO: implement predefined values for fields
# TODO: use validation
# TODO: add is_saved method
# TODO: add load raw data method



"""Setup for kqueen package."""

from setuptools import setup, find_packages

version = '0.15'

with open('README.rst') as f:
    long_description = ''.join(f.readlines())

test_require = [
    'coveralls',
    'faker',
    'flake8',
    'pytest',
    'pytest-cov',
    'pytest-env',
    'pytest-flask',
    'pytest-ordering',
]

setup(
    name='kqueen',
    version=version,
    description='Kubernetes cluster orchestrator',
    long_description=long_description,
    author='Tomáš Kukrál',
    author_email='tomas.kukral@6shore.net',
    license='MIT',
    url='https://github.com/Mirantis/kqueen/',
    download_url='https://github.com/Mirantis/kqueen/archive/v{}.tar.gz'.format(version),
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        'bcrypt',
        'Flask==0.12.2',
        'Flask-JWT==0.3.2',
        'flask-swagger-ui',
        'gunicorn',
        'kubernetes',
        'pycrypto',
        'prometheus_client',
        'python-etcd',
        'python-jenkins',
        'pyyaml',
        'requests',
        'google-api-python-client==1.6.4',
        'google-auth==1.2.1',
        'google-auth-httplib2==0.0.3',
        'azure==2.0.0',
        'azure-mgmt-containerservice',
    ],
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=test_require,
    extras_require={
        'test': test_require,
        'dev': test_require + [
            'ipython',
            'sphinx',
            'sphinx-autobuild',
            'sphinx_rtd_theme',
        ]
    },
    classifiers=[
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        'console_scripts': [
            'kqueen = kqueen.server:run',
        ],
    },
)



from .test_crud import BaseTestCRUD
from flask import url_for
from kqueen.config import current_config
from kqueen.conftest import cluster
from uuid import uuid4

import json
import pytest

config = current_config()


class TestClusterCRUD(BaseTestCRUD):
    def get_object(self):
        obj = cluster()

        return obj

    def get_edit_data(self):
        return {'name': 'patched cluster'}

    def get_create_data(self):
        data = self.obj.get_dict()
        data['id'] = None
        data['provisioner'] = 'Provisioner:{}'.format(self.obj.provisioner.id)
        data['owner'] = 'User:{}'.format(self.obj.owner.id)

        return data

    def test_get_dict_expanded(self):

        dicted = self.obj.get_dict(expand=True)
        assert dicted['provisioner']['id'] == self.obj.provisioner.id

    def test_cluster_get(self):

        cluster_id = self.obj.id

        response = self.client.get(
            url_for('api.cluster_get', pk=cluster_id),
            headers=self.auth_header
        )

        print('JSON response:', response.json, sep='\n')
        print('get_dict:', self.obj.get_dict(expand=True), sep='\n')

        assert response.json == self.obj.get_dict(expand=True)

    @pytest.mark.parametrize('cluster_id,status_code', [
        (uuid4(), 404),
        ('wrong-uuid', 404),
    ])
    def test_object_not_found(self, cluster_id, status_code):
        response = self.client.get(
            url_for('api.cluster_get', pk=cluster_id),
            headers=self.auth_header
        )
        assert response.status_code == status_code

    def test_cluster_status_returns(self):
        cluster_id = self.obj.id

        response = self.client.get(
            url_for('api.cluster_status', pk=cluster_id),
            headers=self.auth_header
        )
        assert response.status_code == 200

        rj = response.json

        assert 'deployments' in rj
        assert 'nodes' in rj
        assert 'nodes_pods' in rj
        assert 'pods' in rj
        assert 'services' in rj
        assert 'version' in rj
        assert 'git_version' in rj['version']
        assert 'platform' in rj['version']

    @pytest.mark.parametrize('cluster_id,status_code', [
        (uuid4(), 404),
        ('sdfsdfabc123', 404),
    ])
    @pytest.mark.parametrize('url', [
        'cluster_status',
        'cluster_kubeconfig',
        'cluster_progress'
    ])
    def test_cluster_status_404(self, url, cluster_id, status_code):
        url = url_for('api.{}'.format(url), pk=cluster_id)
        response = self.client.get(url, headers=self.auth_header)

        assert response.status_code == status_code

    def test_kubeconfig(self):

        url = url_for('api.cluster_kubeconfig', pk=self.obj.id)
        response = self.client.get(url, headers=self.auth_header)
        assert response.json == self.obj.kubeconfig

    def test_topology_data_format(self):

        url = url_for('api.cluster_topology_data', pk=self.obj.id)
        response = self.client.get(url, headers=self.auth_header)

        assert isinstance(response.json, dict)

        assert 'items' in response.json
        assert 'kinds' in response.json
        assert 'relations' in response.json

    def test_progress_format(self):

        url = url_for('api.cluster_progress', pk=self.obj.id)
        response = self.client.get(url, headers=self.auth_header)

        assert isinstance(response.json, dict)

        assert 'response' in response.json
        assert 'progress' in response.json
        assert 'result' in response.json

    def test_create(self, provisioner, user):
        provisioner.save()
        user.save()

        post_data = {
            'name': 'Testing cluster',
            'provisioner': 'Provisioner:{}'.format(provisioner.id),
            'owner': 'User:{}'.format(user.id)
        }

        response = self.client.post(
            url_for('api.cluster_create'),
            data=json.dumps(post_data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 200

        assert 'id' in response.json
        assert response.json['name'] == post_data['name']
        assert response.json['provisioner'] == provisioner.get_dict(expand=True)

    def test_provision_after_create(self, provisioner, user, monkeypatch):
        provisioner.save()
        user.save()

        def fake_provision(self, *args, **kwargs):
            self.cluster.name = 'Provisioned'
            self.cluster.save()

            return True, None

        monkeypatch.setattr(provisioner.get_engine_cls(), 'provision', fake_provision)

        post_data = {
            'name': 'Testing cluster',
            'provisioner': 'Provisioner:{}'.format(provisioner.id),
            'owner': 'User:{}'.format(user.id)
        }

        response = self.client.post(
            url_for('api.cluster_create'),
            data=json.dumps(post_data),
            headers=self.auth_header,
            content_type='application/json',
        )

        object_id = response.json['id']
        obj = self.obj.__class__.load(self.namespace, object_id)

        assert response.status_code == 200
        assert obj.name == 'Provisioned'

    def test_provision_failed(self, provisioner, user, monkeypatch):
        provisioner.save()
        user.save()

        def fake_provision(self, *args, **kwargs):
            return False, 'Testing msg'

        monkeypatch.setattr(provisioner.get_engine_cls(), 'provision', fake_provision)

        post_data = {
            'name': 'Testing cluster',
            'provisioner': 'Provisioner:{}'.format(provisioner.id),
            'owner': 'User:{}'.format(user.id)
        }

        response = self.client.post(
            url_for('api.cluster_create'),
            data=json.dumps(post_data),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 500

    def test_return_400_missing_json(self):
        response = self.client.post(
            url_for('api.cluster_create'),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 400

    @pytest.mark.parametrize('data,code,content_type', [
        (json.dumps({'none': 'anything'}), 500, 'application/json'),
        ('abc', 400, 'text/plain'),
    ])
    def test_error_codes(self, data, code, content_type):
        response = self.client.post(
            url_for('api.cluster_create'),
            data=data,
            headers=self.auth_header,
            content_type=content_type,
        )

        assert response.status_code == code

    def test_cluster_list_run_get_state(self, monkeypatch):
        for _ in range(10):
            c = cluster()
            c.save()

        def fake_get_state(self):
            self.metadata = {'executed': True}
            self.save()

            return config.get('CLUSTER_UNKNOWN_STATE')

        monkeypatch.setattr(self.obj.__class__, 'get_state', fake_get_state)

        response = self.client.get(
            url_for('api.cluster_list'),
            headers=self.auth_header,
            content_type='application/json',
        )

        assert response.status_code == 200

        obj = self.obj.__class__.load(self.namespace, self.obj.id)

        assert obj.metadata, 'get_state wasn\'t executed for cluster {}'.format(obj)
        assert obj.metadata['executed'], 'get_state wasn\'t executed'



from importlib import import_module
from kqueen.config import current_config
from kqueen.kubeapi import KubernetesAPI
from kqueen.storages.etcd import BoolField
from kqueen.storages.etcd import DatetimeField
from kqueen.storages.etcd import IdField
from kqueen.storages.etcd import JSONField
from kqueen.storages.etcd import Model
from kqueen.storages.etcd import ModelMeta
from kqueen.storages.etcd import PasswordField
from kqueen.storages.etcd import RelationField
from kqueen.storages.etcd import StringField
from tempfile import mkstemp

import logging
import os
import subprocess
import yaml

logger = logging.getLogger(__name__)
config = current_config()

#
# Model definition
#


class Cluster(Model, metaclass=ModelMeta):
    id = IdField(required=True)
    name = StringField(required=True)
    provisioner = RelationField()
    state = StringField()
    kubeconfig = JSONField(encrypted=True)
    metadata = JSONField()
    created_at = DatetimeField()
    owner = RelationField(required=True)

    def get_state(self):
        try:
            cluster = self.engine.cluster_get()
        except Exception as e:
            logger.error('Unable to get data from backend for cluster {}'.format(self.name))
            cluster = {}

        if 'state' in cluster:
            if cluster['state'] == self.state:
                return self.state
            self.state = cluster['state']
            self.save()

        return self.state

    @property
    def engine(self):
        if self.provisioner:
            _class = self.provisioner.get_engine_cls()
            if _class:
                parameters = {}
                for i in [self.provisioner.parameters, self.metadata]:
                    if isinstance(i, dict):
                        parameters.update(i)

                return _class(self, **parameters)
        else:
            raise Exception('Missing provisioner')

    def delete(self):
        """Deprovision cluster and delete object from database"""

        deprov_status, deprov_msg = self.engine.deprovision()

        if deprov_status:
            super(Cluster, self).delete()
        else:
            raise Exception('Unable to deprovision cluster: {}'.format(deprov_msg))

    def get_kubeconfig(self):
        if self.kubeconfig:
            return self.kubeconfig
        kubeconfig = self.engine.get_kubeconfig()
        self.kubeconfig = kubeconfig
        self.save()
        return kubeconfig

    def status(self):
        """Return information about Kubernetes cluster"""
        try:
            kubernetes = KubernetesAPI(cluster=self)

            out = {
                'addons': kubernetes.list_services(filter_addons=True),
                'deployments': kubernetes.list_deployments(),
                'namespaces': kubernetes.list_namespaces(),
                'nodes': kubernetes.list_nodes(),
                'nodes_pods': kubernetes.count_pods_by_node(),
                'persistent_volumes': kubernetes.list_persistent_volumes(),
                'persistent_volume_claims': kubernetes.list_persistent_volume_claims(),
                'pods': kubernetes.list_pods(),
                'replica_sets': kubernetes.list_replica_sets(),
                'services': kubernetes.list_services(),
                'version': kubernetes.get_version(),
            }

        except Exception:
            out = {}

        return out

    def topology_data(self):
        """
        Return information about Kubernetes cluster in format used in
        visual processing.
        """
        raw_data = []
        kubernetes = KubernetesAPI(cluster=self)

        nodes = kubernetes.list_nodes()
        for node in nodes:
            node['kind'] = 'Node'

        pods = kubernetes.list_pods(False)
        for pod in pods:
            pod['kind'] = 'Pod'

        namespaces = kubernetes.list_namespaces()
        for namespace in namespaces:
            namespace['kind'] = 'Namespace'

        services = kubernetes.list_services(False)
        for service in services:
            service['kind'] = 'Service'

        deployments = kubernetes.list_deployments(False)
        for deployment in deployments:
            deployment['kind'] = 'Deployment'

        replica_sets = kubernetes.list_replica_sets(False)
        replica_set_dict = {datum['metadata']['uid']: datum for datum in replica_sets}

        raw_data = nodes + pods + services + deployments + namespaces

        resources = {datum['metadata']['uid']: datum for datum in raw_data}
        relations = []

        namespace_name_2_uid = {}
        node_name_2_uid = {}
        service_select_run_2_uid = {}
        service_select_app_2_uid = {}

        for resource_id, resource in resources.items():
            # Add node name to uid mapping
            if resource['kind'] == 'Node':
                node_name_2_uid[resource['metadata']['name']] = resource_id

            # Add node name to uid mapping
            if resource['kind'] == 'Namespace':
                namespace_name_2_uid[resource['metadata']['name']] = resource_id

            # Add service run selector to uid_mapping
            if resource['kind'] == 'Service' and resource['spec'].get('selector', {}) is not None:
                if resource['spec'].get('selector', {}).get('run', False):
                    service_select_run_2_uid[resource['spec']['selector']['run']] = resource_id
                if resource['spec'].get('selector', {}).get('app', False):
                    service_select_app_2_uid[resource['spec']['selector']['app']] = resource_id

            # Add Containers as top-level resource
            """
            if resource['kind'] == 'Pod':
                for container in resource['spec']['containers']:
                    container_id = "{1}-{2}".format(
                        resource['metadata']['uid'], container['name'])
                    resources[container_id] = {
                        'metadata': container,
                        'kind': 'Container'
                    }
                    relations.append({
                        'source': resource_id,
                        'target': container_id,
                    })
            """

        for resource_id, resource in resources.items():
            if resource['kind'] not in ('Node', 'Namespace'):
                relations.append({
                    'source': resource_id,
                    'target': namespace_name_2_uid[resource['metadata']['namespace']]
                })

            if resource['kind'] == 'Pod':

                # define relationship between pods and nodes
                if resource['spec']['node_name'] is not None:
                    relations.append({
                        'source': resource_id,
                        'target': node_name_2_uid[resource['spec']['node_name']]
                    })

                # define relationships between pods and rep sets and
                # replication controllers
                if resource['metadata'].get('owner_references', False):
                    if resource['metadata']['owner_references'][0]['kind'] == 'ReplicaSet':
                        rep_set_id = resource['metadata']['owner_references'][0]['uid']
                        deploy_id = replica_set_dict[rep_set_id]['metadata']['owner_references'][0]['uid']
                        relations.append({
                            'source': deploy_id,
                            'target': resource_id
                        })

                # rel'n between pods and services
                if resource.get('metadata', {}).get('labels', {}).get('run', False):
                    relations.append({
                        'source': resource_id,
                        'target': service_select_run_2_uid[resource['metadata']['labels']['run']]
                    })

                if resource.get('metadata', {}).get('labels', {}).get('app', False):
                    try:
                        relations.append({
                            'source': resource_id,
                            'target': service_select_app_2_uid[resource['metadata']['labels']['app']]
                        })
                    except Exception:
                        pass

        out = {
            'items': resources,
            'relations': relations,
            'kinds': {
                'Pod': '',
            }
        }
#        except Exception:
#            out = {
#                'items': [],
#                'relations': []
#            }

        return out

    def get_kubeconfig_file(self):
        """
        Create file with kubeconfig and make this file available on filesystem.

        Returns:
            str: Filename (including path).

        """

        if hasattr(self, 'kubeconfig_path') and os.path.isfile(self.kubeconfig_path):
            return self.kubeconfig_path

        # create kubeconfig file
        filehandle, file_path = mkstemp()
        filehandle = open(filehandle, 'w')
        filehandle.write(yaml.dump(self.kubeconfig))
        self.kubeconfig_path = file_path

        return file_path

    def apply(self, resource_text):
        """
        Apply YAML file supplied as text

        Args:
            resource_text (text): Content of file to apply

        Returns:
            tuple: (return_code, stdout)


        """
        kubeconfig = self.get_kubeconfig_file()

        # create temporary resource file
        # TODO: create helper for this
        filehandle, file_path = mkstemp()
        filehandle = open(filehandle, 'w')
        filehandle.write(resource_text)
        filehandle.close()

        # apply resource file
        cmd = ['kubectl', '--kubeconfig', kubeconfig, 'apply', '-f', file_path]

        # TODO: validate output
        run = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        return run


class Provisioner(Model, metaclass=ModelMeta):
    id = IdField(required=True)
    name = StringField(required=True)
    verbose_name = StringField(required=False)
    engine = StringField(required=True)
    state = StringField()
    parameters = JSONField(encrypted=True)
    created_at = DatetimeField()
    owner = RelationField(required=True)

    @classmethod
    def list_engines(self):
        """Read engines and filter them according to whitelist"""

        engines = config.get('PROVISIONER_ENGINE_WHITELIST')

        if engines is None:
            from kqueen.engines import __all__ as engines_available
            engines = engines_available

        return engines

    def get_engine_cls(self):
        """Return engine class"""
        try:
            module_path = '.'.join(self.engine.split('.')[:-1])
            class_name = self.engine.split('.')[-1]
            module = import_module(module_path)
            _class = getattr(module, class_name)
        except Exception as e:
            logger.error(repr(e))
            _class = None

        return _class

    def engine_status(self, save=True):
        state = config.get('PROVISIONER_UNKNOWN_STATE')
        engine_class = self.get_engine_cls()
        if engine_class:
            state = engine_class.engine_status()
        if save:
            self.state = state
            self.save()
        return state

    def alive(self):
        """Test availability of provisioner and return bool"""
        return True

    def save(self, check_status=True):
        if check_status:
            self.state = self.engine_status(save=False)
        self.verbose_name = getattr(self.get_engine_cls(), 'verbose_name', self.engine)
        return super(Provisioner, self).save()


#
# AUTHENTICATION
#


class Organization(Model, metaclass=ModelMeta):
    global_namespace = True

    id = IdField(required=True)
    name = StringField(required=True)
    namespace = StringField(required=True)
    policy = JSONField()
    created_at = DatetimeField()


class User(Model, metaclass=ModelMeta):
    global_namespace = True

    id = IdField(required=True)
    username = StringField(required=True)
    email = StringField(required=False)
    password = PasswordField(required=True)
    organization = RelationField(required=True)
    created_at = DatetimeField()
    role = StringField(required=True)
    active = BoolField(required=True)
    metadata = JSONField(required=False)

    @property
    def namespace(self):
        """
        Get namespace from organization.

        Returns:
            str: Namespace (from organization)
        """

        return self.organization.namespace
